from typing import Union

from merakicommons.container import SearchableList
from merakicommons.cache import lazy_property

from ...configuration import settings
from ...data import Region, Platform
from ...dto.staticdata.version import VersionListDto
from ..common import CoreData, DataObjectList, CassiopeiaGhostList


class VersionListData(DataObjectList):
    _dto_type = VersionListDto
    _renamed = {}

    @property
    def region(self) -> str:
        return self._dto["region"]


class Versions(CassiopeiaGhostList):
    _data_types = {VersionListData}

    def __init__(self, *args, region: Union[Region, str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        kwargs = {"region": region}
        super().__init__(*args, **kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        self.clear()
        SearchableList.__init__(self, [v for v in data])
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[VersionListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform



from typing import Type, TypeVar, MutableMapping, Any, Iterable, Generator
from io import BytesIO

from datapipelines import DataSource, PipelineContext, Query, NotFoundError
from PIL import Image as ImageLoader
from PIL.Image import Image

from .common import HTTPClient

T = TypeVar("T")


class ImageDataSource(DataSource):
    def __init__(self, http_client: HTTPClient = None) -> None:
        if http_client is None:
            self._client = HTTPClient()
        else:
            self._client = http_client

    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    _validate_get_image = Query. \
        has("url").as_(str)

    @get.register(Image)
    def get_image(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Image:
        ImageDataSource._validate_get_image(query, context)
        try:
            data, _ = self._client.get(query["url"])
            return ImageLoader.open(BytesIO(data))
        except Exception as e:
            raise NotFoundError(str(e)) from e

    _validate_get_many_image = Query. \
        has("urls").as_(Iterable)

    @get_many.register(Image)
    def get_many_image(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[Image, None, None]:
        ImageDataSource._validate_get_many_image(query, context)

        def generator():
            for url in query["urls"]:
                if not isinstance(url, str):
                    raise NotFoundError("urls must be strings!")

                try:
                    data, _ = self._client.get(url)
                    yield ImageLoader.open(BytesIO(data))
                except Exception as e:
                    raise NotFoundError(str(e)) from e

        return generator()



from .common import CoreData
from ..dto import champion as dto


class ChampionListData(list):
    _dto_type = dto.ChampionListDto


class ChampionData(CoreData):
    _dto_type = dto.ChampionDto
    _renamed = {"ranked_enabled": "rankedPlayEnabled", "custom_enabled": "botEnabled", "coop_ai_enabled": "botMmEnabled",
                "enabled": "active", "free_to_play": "freeToPlay", "champ_data": "champData"}

    @property
    def enabled(self) -> bool:
        return self._dto["active"]

    @property
    def custom_enabled(self) -> bool:
        return self._dto["botEnabled"]

    @property
    def coop_ai_enabled(self) -> bool:
        return self._dto["botMmEnabled"]

    @property
    def free_to_play(self) -> bool:
        return self._dto["freeToPlay"]

    @property
    def id(self) -> int:
        return self._dto["id"]

    @property
    def ranked_enabled(self) -> bool:
        return self._dto["rankedPlayEnabled"]

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def platform(self) -> str:
        return self._dto["platform"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]

    @property
    def champ_data(self) -> set:
        return self._dto["champData"]


# No non-CoreData core type is needed



import cassiopeia
import pytest

from merakicommons.container import SearchableList, SearchError

from .constants import CHAMP_NAME


def test_returns_correct_type():
    champ = cassiopeia.get_champion(CHAMP_NAME)
    all_champs = cassiopeia.get_champions()

    assert isinstance(champ, cassiopeia.Champion)
    assert isinstance(all_champs, SearchableList)
    assert all(isinstance(c, cassiopeia.Champion) for c in all_champs)


def test_raises_with_unknown_champion():
    with pytest.raises(SearchError):
        cassiopeia.get_champion('nonexistant champ')


def test_champion_and_champions_return_same_data():
    champions = cassiopeia.get_champions()

    champ = champions[0]
    from_get_champion = cassiopeia.get_champion(champ.name)

    assert champ.__dict__ == from_get_champion.__dict__



from typing import List, Mapping, Union

from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable, SearchableList, SearchableDictionary

from ..data import Region, Platform
from .common import CoreData, DataObjectList, CassiopeiaGhost, CassiopeiaGhostList
from .summoner import Summoner
from ..dto.masterypage import MasteryDto, MasteryPageDto, MasteryPagesDto
from .staticdata.mastery import Mastery as StaticdataMastery


##############
# Data Types #
##############


class MasteryPagesData(DataObjectList):
    _dto_type = MasteryPagesDto
    _renamed = {}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def summoner_id(self) -> str:
        return self._dto["summonerId"]


class MasteryData(CoreData):
    """This object contains mastery information."""
    _dto_type = MasteryDto
    _renamed = {"points": "rank"}

    @property
    def id(self) -> int:
        """Mastery ID. For static information correlating to masteries, please refer to the LoL Static Data API."""
        return self._dto["id"]

    @property
    def points(self) -> int:
        """Mastery rank (i.e., the number of points put into this mastery)."""
        return self._dto["rank"]


class MasteryPageData(CoreData):
    """This object contains mastery page information."""
    _dto_type = MasteryPageDto
    _renamed = {}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def current(self) -> bool:
        """Indicates if the mastery page is the current mastery page."""
        return self._dto["current"]

    @property
    def masteries(self) -> List[MasteryData]:
        """Collection of masteries associated with the mastery page."""
        return [MasteryData.from_dto(mastery) for mastery in self._dto["masteries"]]

    @property
    def name(self) -> str:
        """Mastery page name."""
        return self._dto["name"]

    @property
    def id(self) -> int:
        """Mastery page ID."""
        return self._dto["id"]


##############
# Core Types #
##############


class MasteryPages(CassiopeiaGhostList):
    _data_types = {MasteryPagesData}

    def __init__(self, *args, summoner: Union[Summoner, int, str], region: Union[Region, str] = None):
        super().__init__(*args, region=region)
        if isinstance(summoner, str):
            summoner = Summoner(name=summoner)
        elif isinstance(summoner, int):
            summoner = Summoner(id=summoner)
        self.__summoner = summoner

    def __get_query__(self):
        return {"summoner.id": self.__summoner.id, "region": self.region, "platform": self.platform}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        self.clear()
        from ..transformers.masteries import MasteriesTransformer
        SearchableList.__init__(self, [MasteriesTransformer.mastery_page_data_to_core(None, i) for i in data])
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[MasteryPagesData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform


@searchable({str: ["name", "masteries", "region", "platform", "locale"], int: ["id"], bool: ["current"], StaticdataMastery: ["masteries"], Region: ["region"], Platform: ["platform"]})
class MasteryPage(CassiopeiaGhost):
    _data_types = {MasteryPageData}
    _load_types = {MasteryPageData: MasteryPages}

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[MasteryPageData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this mastery page."""
        return self.region.platform

    @property
    def locale(self) -> str:
        """The locale for this mastery page."""
        return self._data[MasteryPageData].locale

    @lazy_property
    def summoner(self) -> Summoner:
        return Summoner(id=self._data[MasteryPageData].summoner_id)

    @CassiopeiaGhost.property(MasteryPageData)
    @ghost_load_on(KeyError)
    def current(self) -> bool:
        return self._data[MasteryPageData].current

    @CassiopeiaGhost.property(MasteryPageData)
    @ghost_load_on(KeyError)
    @lazy
    def masteries(self) -> Mapping[StaticdataMastery, int]:
        return SearchableDictionary({StaticdataMastery(id=mastery.id, region=self.region.value): mastery.points for mastery in self._data[MasteryPageData].masteries})

    @CassiopeiaGhost.property(MasteryPageData)
    @ghost_load_on(KeyError)
    def name(self) -> str:
        return self._data[MasteryPageData].name

    @CassiopeiaGhost.property(MasteryPageData)
    @ghost_load_on(KeyError)
    def id(self) -> int:
        return self._data[MasteryPageData].id



import os
from PIL.Image import Image as PILImage
from typing import Union

from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy_property, lazy
from merakicommons.container import searchable, SearchableList

from ...configuration import settings
from ...data import Region, Platform
from ...dto.staticdata.profileicon import ProfileIconDataDto, ProfileIconListDto
from ..common import CoreData, DataObjectList, CassiopeiaGhost, CassiopeiaGhostList, get_latest_version


try:
    import ujson as json
except ImportError:
    import json

_profile_icon_names = None


##############
# Data Types #
##############


class ProfileIconListData(DataObjectList):
    _dto_type = ProfileIconListDto
    _renamed = {}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]


class ProfileIconData(CoreData):
    _dto_type = ProfileIconDataDto
    _renamed = {"id": "profileIconId"}

    @property
    def id(self) -> int:
        return self._dto["profileIconId"]

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]


##############
# Core Types #
##############


class ProfileIcons(CassiopeiaGhostList):
    _data_types = {ProfileIconData}

    def __init__(self, *args, region: Union[Region, str] = None, version: str = None, locale: str = None):
        if region is None:
            region = settings.default_region
        kwargs = {"region": region}
        if version is not None:
            kwargs["version"] = version
        if locale is not None:
            kwargs["locale"] = locale
        super().__init__(*args, **kwargs)

    def __get_query__(self):
        query = {"region": self.region, "platform": self.platform, "version": self.version}
        try:
            query["locale"] = self.locale
        except KeyError:
            pass
        return query

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        self.clear()
        from ...transformers.staticdata import StaticDataTransformer
        SearchableList.__init__(self, [StaticDataTransformer.profile_icon_data_to_core(None, i) for i in data])
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[ProfileIconData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def version(self) -> str:
        try:
            return self._data[ProfileIconData].version
        except KeyError:
            version = get_latest_version(region=self.region, endpoint="profileicon")
            self(version=version)
            return self._data[ProfileIconData].version

    @property
    def locale(self) -> str:
        return self._data[ProfileIconData].locale


@searchable({int: ["id"], str: ["name", "url"], PILImage: ["image"]})
class ProfileIcon(CassiopeiaGhost):
    _data_types = {ProfileIconData}
    _load_types = {ProfileIconData: ProfileIconListData}

    def __init__(self, *, id: int, region: Union[Region, str] = None, version: str = None, locale: str = None):
        if region is None:
            region = settings.default_region
        kwargs = {"id": id, "region": region}
        if version is not None:
            kwargs["version"] = version
        if locale is not None:
            kwargs["locale"] = locale
        super().__init__(**kwargs)

    def __get_query__(self):
        query = {"region": self.region, "platform": self.platform, "version": self.version}
        try:
            query["locale"] = self.locale
        except KeyError:
            pass
        return query

    def __load_hook__(self, load_group, data) -> None:
        def find_matching_attribute(datalist, attrname, attrvalue):
            for item in datalist:
                if getattr(item, attrname, None) == attrvalue:
                    return item
        data = find_matching_attribute(data, "id", self.id)
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        """The region for this profile icon."""
        return Region(self._data[ProfileIconData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this profile icon."""
        return self.region.platform

    @property
    def version(self) -> str:
        """The version for this profile icon."""
        try:
            return self._data[ProfileIconData].version
        except KeyError:
            version = get_latest_version(region=self.region, endpoint="profileicon")
            self(version=version)
            return self._data[ProfileIconData].version

    @property
    def locale(self) -> str:
        """The locale for this profile icon."""
        return self._data[ProfileIconData].locale

    @property
    def id(self) -> int:
        return self._data[ProfileIconData].id

    @lazy_property
    def name(self) -> Union[str, None]:
        global _profile_icon_names
        if _profile_icon_names is None:
            module_directory = os.path.dirname(os.path.realpath(__file__))
            module_directory, _ = os.path.split(module_directory)  # Go up one directory
            module_directory, _ = os.path.split(module_directory)  # Go up another directory
            filename = os.path.join(module_directory, "profile_icon_names.json")
            _profile_icon_names = json.load(open(filename))
            _profile_icon_names = {int(key): value for key, value in _profile_icon_names.items()}
        try:
            return _profile_icon_names[self._data[ProfileIconData].id]
        except KeyError:
            return None

    @CassiopeiaGhost.property(ProfileIconData)
    @ghost_load_on(KeyError)
    def url(self) -> str:
        version = get_latest_version(region=self.region, endpoint="profileicon")
        return "http://ddragon.leagueoflegends.com/cdn/{version}/img/profileicon/{id}.png".format(version=version, id=self.id)

    @CassiopeiaGhost.property(ProfileIconData)
    @ghost_load_on(KeyError)
    @lazy
    def image(self) -> PILImage:
        return settings.pipeline.get(PILImage, query={"url": self.url})



from typing import Dict, Union

from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy_property
from merakicommons.container import searchable

from ...configuration import settings
from ...data import Region, Platform
from ..common import CoreData, CassiopeiaGhost, get_latest_version
from ...dto.staticdata import realm as dto


##############
# Data Types #
##############


class LanguageStringsData(CoreData):
    _dto_type = dto.RealmDto
    _renamed = {"strings": "data"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]

    @property
    def type(self) -> str:
        return self._dto["type"]

    @property
    def strings(self) -> Dict[str, str]:
        return self._dto["data"]


##############
# Core Types #
##############


@searchable({})
class LanguageStrings(CassiopeiaGhost):
    _data_types = {LanguageStringsData}

    def __init__(self, *, strings: Dict[str, str] = None, region: Union[Region, str] = None, version: str = None, locale: str = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        if locale is None:
            locale = region.default_locale
        kwargs = {"region": region, "locale": locale}
        if version is not None:
            kwargs["version"] = version
        if strings is not None:
            kwargs["strings"] = strings
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale}

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[LanguageStringsData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def version(self) -> str:
        try:
            return self._data[LanguageStringsData].version
        except KeyError:
            version = get_latest_version(region=self.region, endpoint="language")
            self(version=version)
            return self._data[LanguageStringsData].version

    @lazy_property
    def locale(self) -> str:
        return self._data[LanguageStringsData].locale

    @CassiopeiaGhost.property(LanguageStringsData)
    @ghost_load_on(KeyError)
    def type(self) -> str:
        return self._data[LanguageStringsData].type

    @CassiopeiaGhost.property(LanguageStringsData)
    @ghost_load_on(KeyError)
    def strings(self) -> Dict[str, str]:
        return self._data[LanguageStringsData].strings



from PIL.Image import Image as PILImage

from merakicommons.container import searchable
from merakicommons.cache import lazy_property
from PIL.Image import Image as PILImage

from ...configuration import settings
from ..common import CoreData, CassiopeiaObject


class SpriteData(CoreData):
    _renamed = {"height": "h", "width": "w"}

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def sprite(self) -> str:
        return self._dto["sprite"]

    @property
    def x(self) -> int:
        return self._dto["x"]

    @property
    def y(self) -> int:
        return self._dto["y"]

    @property
    def width(self) -> int:
        return self._dto["w"]

    @property
    def height(self) -> int:
        return self._dto["h"]


class ImageData(CoreData):
    _renamed = {"height": "h", "width": "w"}

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def full(self) -> str:
        return self._dto["full"]

    @property
    def group(self) -> str:
        return self._dto["group"]

    @property
    def height(self) -> int:
        return self._dto["h"]

    @property
    def width(self) -> int:
        return self._dto["w"]

    @property
    def y(self) -> int:
        return self._dto["y"]

    @property
    def x(self) -> int:
        return self._dto["x"]

    @property
    def sprite(self) -> str:
        return self._dto["sprite"]


@searchable({str: ["sprite", "url"]})
class Sprite(CassiopeiaObject):
    _data_types = {SpriteData}

    @property
    def version(self) -> str:
        return self._data[SpriteData].version

    @property
    def sprite(self) -> str:
        return self._data[SpriteData].sprite

    @property
    def y(self) -> int:
        return self._data[SpriteData].y

    @property
    def x(self) -> int:
        return self._data[SpriteData].x

    @property
    def width(self) -> int:
        return self._data[SpriteData].width

    @property
    def height(self) -> int:
        return self._data[SpriteData].height

    @property
    def url(self) -> str:
        return "http://ddragon.leagueoflegends.com/cdn/{version}/img/sprite/{sprite}".format(version=self.version, sprite=self.sprite)

    @lazy_property
    def image(self) -> PILImage:
        return settings.pipeline.get(PILImage, query={"url": self.url})


@searchable({str: ["full", "url"]})
class Image(CassiopeiaObject):
    _data_types = {ImageData}

    @property
    def version(self) -> str:
        return self._data[ImageData].version

    @property
    def full(self) -> str:
        return self._data[ImageData].full

    @property
    def group(self) -> str:
        return self._data[ImageData].group

    @property
    def url(self) -> str:
        return "http://ddragon.leagueoflegends.com/cdn/{version}/img/{group}/{full}".format(version=self.version, group=self.group, full=self.full)

    @lazy_property
    def image(self) -> PILImage:
        return settings.pipeline.get(PILImage, query={"url": self.url})

    @lazy_property
    def sprite_info(self) -> Sprite:
        sprite = Sprite(w=self._data[ImageData].width,
                        h=self._data[ImageData].height,
                        x=self._data[ImageData].x,
                        y=self._data[ImageData].y,
                        sprite=self._data[ImageData].sprite,
                        version=self._data[ImageData].version)
        return sprite



import pytest

from cassiopeia.patches import patches, Patch


def test_known_patches():
    assert isinstance(Patch.from_str('0.9.22.4'), Patch)
    assert isinstance(Patch.from_str('5.19'), Patch)


def test_unknown_patch_raises():
    with pytest.raises(ValueError):
        Patch.from_str("unknown patch")


def test_patches_sorted():
    assert sorted(patches, key=lambda p: p.start) == patches


def test_major():
    assert patches[0].major == '0'
    assert Patch.from_str('6.10').major == '6'


def test_minor():
    assert patches[0].minor == '9'
    assert Patch.from_str('6.10').minor == '10'


def test_majorminor():
    assert patches[0].majorminor == '0.9'
    assert Patch.from_str('6.10').majorminor == '6.10'


def test_revision():
    assert patches[0].revision == '22.4'
    assert Patch.from_str('6.10').revision == ''




from typing import List, Set, Mapping, Union

from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable, SearchableList

from ...configuration import settings
from ...data import Region, Platform, Map
from ..common import CoreData, CassiopeiaObject, CassiopeiaGhost, CassiopeiaGhostList, DataObjectList, get_latest_version
from .common import ImageData, Sprite, Image
from ...dto.staticdata import item as dto


##############
# Data Types #
##############


class ItemListData(DataObjectList):
    _dto_type = dto.ItemListDto
    _renamed = {"included_data": "includedData"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]

    @property
    def included_data(self) -> Set[str]:
        return self._dto["includedData"]


class ItemTreeData(CoreData):
    _renamed = {}

    @property
    def header(self) -> str:
        return self._dto["header"]

    @property
    def tags(self) -> List[str]:
        return self._dto["tags"]


class ItemStatsData(CoreData):
    _renamed = {"percent_critical_strike_damage": "PercentCritDamageMod", "percent_magic_resist": "PercentSpellBlockMod", "percent_health_regen": "PercentHPRegenMod", "percent_movespeed": "PercentMovementSpeedMod", "magic_resist": "FlatSpellBlockMod", "critical_strike_damage": "FlatCritDamageMod", "energy": "FlatEnergyPoolMod", "life_steal": "PercentLifeStealMod", "mana": "FlatMPPoolMod", "movespeed": "FlatMovementSpeedMod", "percent_attack_speed": "PercentAttackSpeedMod", "block": "FlatBlockMod", "percent_block": "PercentBlockMod", "energy_regen": "FlatEnergyRegenMod", "spell_vamp": "PercentSpellVampMod", "mana_regen": "FlatMPRegenMod", "dodge": "PercentDodgeMod", "attack_speed": "FlatAttackSpeedMod", "armor": "FlatArmorMod", "health_regen": "FlatHPRegenMod", "percent_ability_power": "PercentMagicDamageMod", "percent_mana_regen": "PercentMPPoolMod", "ability_power": "FlatMagicDamageMod", "percent_mana_regen": "PercentMPRegenMod", "percent_attack_damage": "PercentPhysicalDamageMod", "attack_damage": "FlatPhysicalDamageMod", "percent_health": "PercentHPPoolMod", "percent_armor": "PercentArmorMod", "percent_xp_bonus": "PercentEXPBonus", "health": "FlatHPPoolMod", "critical_strike_chance": "FlatCritChanceMod", "xp_bonus": "FlatEXPBonus"}

    @property
    def percent_critical_strike_damage(self) -> float:
        return self._dto["PercentCritDamageMod"]

    @property
    def percent_magic_resist(self) -> float:
        return self._dto["PercentSpellBlockMod"]

    @property
    def percent_health_regen(self) -> float:
        return self._dto["PercentHPRegenMod"]

    @property
    def percent_movespeed(self) -> float:
        return self._dto["PercentMovementSpeedMod"]

    @property
    def magic_resist(self) -> str:
        return self._dto["FlatSpellBlockMod"]

    @property
    def critical_strike_damage(self) -> float:
        return self._dto["FlatCritDamageMod"]

    @property
    def energy(self) -> float:
        return self._dto["FlatEnergyPoolMod"]

    @property
    def life_steal(self) -> float:
        return self._dto["PercentLifeStealMod"]

    @property
    def mana(self) -> float:
        return self._dto["FlatMPPoolMod"]

    @property
    def movespeed(self) -> float:
        return self._dto["FlatMovementSpeedMod"]

    @property
    def percent_attack_speed(self) -> float:
        return self._dto["PercentAttackSpeedMod"]

    @property
    def block(self) -> float:
        return self._dto["FlatBlockMod"]

    @property
    def percent_block(self) -> float:
        return self._dto["PercentBlockMod"]

    @property
    def energy_regen(self) -> float:
        return self._dto["FlatEnergyRegenMod"]

    @property
    def spell_vamp(self) -> float:
        return self._dto["PercentSpellVampMod"]

    @property
    def mana_regen(self) -> float:
        return self._dto["FlatMPRegenMod"]

    @property
    def dodge(self) -> float:
        return self._dto["PercentDodgeMod"]

    @property
    def attack_speed(self) -> float:
        return self._dto["FlatAttackSpeedMod"]

    @property
    def armor(self) -> float:
        return self._dto["FlatArmorMod"]

    @property
    def health_regen(self) -> float:
        return self._dto["FlatHPRegenMod"]

    @property
    def percent_ability_power(self) -> float:
        return self._dto["PercentMagicDamageMod"]

    @property
    def percent_mana_regen(self) -> float:
        return self._dto["PercentMPPoolMod"]

    @property
    def ability_power(self) -> float:
        return self._dto["FlatMagicDamageMod"]

    @property
    def percent_mana_regen(self) -> float:
        return self._dto["PercentMPRegenMod"]

    @property
    def percent_attack_damage(self) -> float:
        return self._dto["PercentPhysicalDamageMod"]

    @property
    def attack_damage(self) -> float:
        return self._dto["FlatPhysicalDamageMod"]

    @property
    def percent_health(self) -> float:
        return self._dto["PercentHPPoolMod"]

    @property
    def percent_armor(self) -> float:
        return self._dto["PercentArmorMod"]

    @property
    def percent_xp_bonus(self) -> float:
        return self._dto["PercentEXPBonus"]

    @property
    def health(self) -> float:
        return self._dto["FlatHPPoolMod"]

    @property
    def critical_strike_chance(self) -> float:
        return self._dto["FlatCritChanceMod"] + self._dto["PercentCritChanceMod"]

    @property
    def xp_bonus(self) -> float:
        return self._dto["FlatEXPBonus"]


class GoldData(CoreData):
    _renamed = {}

    @property
    def sell(self) -> int:
        return self._dto["sell"]

    @property
    def total(self) -> int:
        return self._dto["total"]

    @property
    def base(self) -> int:
        return self._dto["base"]

    @property
    def purchaseable(self) -> bool:
        return self._dto["purchaseable"]


class ItemData(CoreData):
    _dto_type = dto.ItemDto
    _renamed = {"hide": "hideFromAll", "in_store": "inStore", "builds_into": "into", "builds_from": "from", "keywords": "colloq", "champion": "requiredChampion", "consume_on_full": "consumeOnFull", "sanitized_description": "sanitizedDescription", "tier": "depth", "max_stacks": "stacks", "included_data": "includedData"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]

    @property
    def included_data(self) -> Set[str]:
        return self._dto["includedData"]

    @property
    def gold(self) -> GoldData:
        return self._dto["gold"]

    @property
    def plaintext(self) -> str:
        return self._dto["plaintext"]

    @property
    def hide(self) -> bool:
        return self._dto["hideFromAll"]

    @property
    def in_store(self) -> bool:
        return self._dto["inStore"]

    @property
    def builds_into(self) -> List[int]:  # TODO Convert str to int in transformer
        return self._dto["into"]

    @property
    def builds_from(self) -> List[int]:  # TODO Convert str to int in transformer
        return self._dto["from"]

    @property
    def id(self) -> int:
        return self._dto["id"]

    @property
    def stats(self) -> ItemStatsData:
        return ItemStatsData.from_dto(self._dto["stats"])

    @property
    def keywords(self) -> Set[str]:
        return set(kw for kw in self._dto["colloq"].split(";") if kw != "")

    @property
    def maps(self) -> List[Map]:  # TODO Convert from Dict to List
        """List of maps where this item is available."""
        return self._dto["maps"]

    @property
    def special_recipe(self) -> int:
        return self._dto["specialRecipe"]

    @property
    def image(self) -> ImageData:
        return ImageData.from_dto(self._dto["image"])

    @property
    def description(self) -> str:
        return self._dto["description"]

    @property
    def tags(self) -> List[str]:
        return self._dto["tags"]

    @property
    def effect(self) -> Mapping[str, str]:
        return self._dto["effect"]

    @property
    def champion(self) -> str:
        return self._dto["requiredChampion"]

    @property
    def group(self) -> str:
        return self._dto["group"]

    @property
    def name(self) -> str:
        return self._dto["name"]

    @property
    def consume_on_full(self) -> bool:
        return self._dto["consumeOnFull"]

    @property
    def consumed(self) -> bool:
        return self._dto["consumed"]

    @property
    def sanitized_description(self) -> str:
        return self._dto["sanitizedDescription"]

    @property
    def tier(self) -> int:
        return self._dto["depth"]

    @property
    def max_stacks(self) -> int:
        return self._dto["stacks"]


##############
# Core Types #
##############


class Items(CassiopeiaGhostList):
    _data_types = {ItemListData}

    def __init__(self, *args, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        if included_data is None:
            included_data = {"all"}
        if locale is None:
            locale = region.default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if version:
            kwargs["version"] = version
        super().__init__(*args, **kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale, "includedData": self.included_data}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        self.clear()
        from ...transformers.staticdata import StaticDataTransformer
        SearchableList.__init__(self, [StaticDataTransformer.item_data_to_core(None, i) for i in data])
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[ItemListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def version(self) -> str:
        try:
            return self._data[ItemListData].version
        except KeyError:
            version = get_latest_version(region=self.region, endpoint="item")
            self(version=version)
            return self._data[ItemListData].version

    @property
    def locale(self) -> str:
        """The locale for this item."""
        return self._data[ItemListData].locale

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additional information for this item when it's loaded."""
        return self._data[ItemListData].included_data


class ItemStats(CassiopeiaObject):
    _data_types = {ItemStatsData}

    @property
    def percent_critical_strike_damage(self) -> float:
        return self._data[ItemStatsData].percent_critical_strike_damage

    @property
    def percent_magic_resist(self) -> float:
        return self._data[ItemStatsData].percent_magic_resist

    @property
    def percent_health_regen(self) -> float:
        return self._data[ItemStatsData].percent_health_regen

    @property
    def percent_movespeed(self) -> float:
        return self._data[ItemStatsData].percent_movespeed

    @property
    def magic_resist(self) -> float:
        return self._data[ItemStatsData].magic_resist

    @property
    def critical_strike_damage(self) -> float:
        return self._data[ItemStatsData].critical_strike_damage

    @property
    def energy(self) -> float:
        return self._data[ItemStatsData].energy

    @property
    def life_steal(self) -> float:
        return self._data[ItemStatsData].life_steal

    @property
    def mana(self) -> float:
        return self._data[ItemStatsData].mana

    @property
    def movespeed(self) -> float:
        return self._data[ItemStatsData].movespeed

    @property
    def percent_attack_speed(self) -> float:
        return self._data[ItemStatsData].percent_attack_speed

    @property
    def block(self) -> float:
        return self._data[ItemStatsData].block

    @property
    def percent_block(self) -> float:
        return self._data[ItemStatsData].percent_block

    @property
    def energy_regen(self) -> float:
        return self._data[ItemStatsData].energy_regen

    @property
    def spell_vamp(self) -> float:
        return self._data[ItemStatsData].spell_vamp

    @property
    def mana_regen(self) -> float:
        return self._data[ItemStatsData].mana_regen

    @property
    def dodge(self) -> float:
        return self._data[ItemStatsData].dodge

    @property
    def attack_speed(self) -> float:
        return self._data[ItemStatsData].attack_speed

    @property
    def armor(self) -> float:
        return self._data[ItemStatsData].armor

    @property
    def health_regen(self) -> float:
        return self._data[ItemStatsData].health_regen

    @property
    def percent_ability_power(self) -> float:
        return self._data[ItemStatsData].percent_ability_power

    @property
    def percent_mana_regen(self) -> float:
        return self._data[ItemStatsData].percent_mana_regen

    @property
    def ability_power(self) -> float:
        return self._data[ItemStatsData].ability_power

    @property
    def percent_mana_regen(self) -> float:
        return self._data[ItemStatsData].percent_mana_regen

    @property
    def percent_attack_damage(self) -> float:
        return self._data[ItemStatsData].percent_attack_damage

    @property
    def attack_damage(self) -> float:
        return self._data[ItemStatsData].attack_damage

    @property
    def percent_health(self) -> float:
        return self._data[ItemStatsData].percent_health

    @property
    def percent_armor(self) -> float:
        return self._data[ItemStatsData].percent_armor

    @property
    def percent_xp_bonus(self) -> float:
        return self._data[ItemStatsData].percent_xp_bonus

    @property
    def health(self) -> float:
        return self._data[ItemStatsData].health

    @property
    def critical_strike_chance(self) -> float:
        return self._data[ItemStatsData].critical_strike_chance

    @property
    def xp_bonus(self) -> float:
        return self._data[ItemStatsData].xp_bonus


class Gold(CassiopeiaObject):
    _data_types = {GoldData}

    @property
    def sell(self) -> int:
        return self._data[GoldData].sell

    @property
    def total(self) -> int:
        return self._data[GoldData].total

    @property
    def base(self) -> int:
        return self._data[GoldData].base

    @property
    def purchaseable(self) -> bool:
        return self._data[GoldData].purchaseable


@searchable({str: ["name", "region", "platform", "locale", "keywords", "maps", "tags", "tier"], int: ["id"], Region: ["region"], Platform: ["platform"], Map: ["maps"]})
class Item(CassiopeiaGhost):
    _data_types = {ItemData}
    _load_types = {ItemData: ItemListData}

    def __init__(self, *, id: int = None, name: str = None, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        if included_data is None:
            included_data = {"all"}
        if locale is None:
            locale = region.default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if id is not None:
            kwargs["id"] = id
        if name is not None:
            kwargs["name"] = name
        if version is not None:
            kwargs["version"] = version
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale, "includedData": self.included_data}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        def find_matching_attribute(datalist, attrname, attrvalue):
            for item in datalist:
                if getattr(item, attrname, None) == attrvalue:
                    return item

        # The `data` is a dict of summoner spell data instances
        if "id" in self._data[ItemData]._dto:
            find = "id", self.id
        elif "name" in self._data[ItemData]._dto:
            find = "name", self.name
        else:
            raise RuntimeError("Impossible!")
        data = find_matching_attribute(data, *find)
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        """The region for this item."""
        return Region(self._data[ItemData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this item."""
        return self.region.platform

    @property
    def version(self) -> str:
        """The version for this item."""
        try:
            return self._data[ItemData].version
        except KeyError:
            version = get_latest_version(region=self.region, endpoint="item")
            self(version=version)
            return self._data[ItemData].version

    @property
    def locale(self) -> str:
        """The locale for this item."""
        return self._data[ItemData].locale

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additonal information for this item when it's loaded."""
        return self._data[ItemData].included_data

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def id(self) -> int:
        """The item's ID."""
        return self._data[ItemData].id

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def gold(self) -> Gold:
        return self._data[ItemData].gold

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def plaintext(self) -> str:
        return self._data[ItemData].plaintext

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def hide(self) -> bool:
        return self._data[ItemData].hide

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def in_store(self) -> bool:
        return self._data[ItemData].in_store

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def builds_into(self) -> List["Item"]:
        return self._data[ItemData].builds_into

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def builds_from(self) -> List["Item"]:
        return self._data[ItemData].builds_from

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def stats(self) -> ItemStats:
        return self._data[ItemData].stats

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def keywords(self) -> List[str]:
        return SearchableList(self._data[ItemData].keywords)

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def maps(self) -> List[Map]:
        return self._data[ItemData].maps

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def description(self) -> str:
        return self._data[ItemData].description

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def tags(self) -> List[str]:
        return self._data[ItemData].tags

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def effect(self) -> Mapping[str, str]:
        return self._data[ItemData].effect

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def champion(self) -> "Champion":
        from .champion import Champion
        return Champion(name=self._data[ItemData].champion)

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def group(self) -> str:
        return self._data[ItemData].group

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def name(self) -> str:
        return self._data[ItemData].name

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def consume_on_full(self) -> bool:
        return self._data[ItemData].consume_on_full

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def consumed(self) -> bool:
        return self._data[ItemData].consumed

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def sanitized_description(self) -> str:
        return self._data[ItemData].sanitized_description

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def tier(self) -> int:
        return self._data[ItemData].tier

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def max_stacks(self) -> int:
        return self._data[ItemData].max_stacks

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    def special_recipe(self) -> int:
        return self._data[ItemData].special_recipe

    @CassiopeiaGhost.property(ItemData)
    @ghost_load_on(KeyError)
    @lazy
    def image(self) -> Image:
        """The image information for this item."""
        image = Image.from_data(self._data[ItemData].image)
        image(version=self.version)
        return image

    @lazy_property
    def sprite(self) -> Sprite:
        return self.image.sprite_info



from .common import DtoObject


class LeaguesListDto(DtoObject):
    pass


class LeagueListDto(DtoObject):
    pass


class LeagueItemDto(DtoObject):
    pass


class MiniSeriesDto(DtoObject):
    pass


class LeaguePositionsDto(DtoObject):
    pass


class ChallengerLeagueListDto(DtoObject):
    pass


class MasterLeagueListDto(DtoObject):
    pass



from typing import List, Mapping, Union
from collections import Counter

from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable, SearchableList, SearchableDictionary

from ..data import Region, Platform
from .common import CoreData, DataObjectList, CassiopeiaGhost, CassiopeiaGhostList
from .summoner import Summoner
from ..dto.runepage import RuneSlotDto, RunePageDto, RunePagesDto
from .staticdata.rune import Rune as StaticdataRune


##############
# Data Types #
##############


class RunePagesData(DataObjectList):
    _dto_type = RunePagesDto
    _renamed = {}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def summoner_id(self) -> str:
        return self._dto["summonerId"]


class RuneSlotData(CoreData):
    """This object contains rune slot information."""
    _dto_type = RuneSlotDto
    _renamed = {"id": "runeId", "slot": "runeSlotId"}

    @property
    def id(self) -> int:
        """Rune ID associated with the rune slot. For static information correlating to rune IDs, please refer to the LoL Static Data API."""
        return self._dto["runeId"]

    @property
    def slot(self) -> int:
        """Rune slot ID."""
        return self._dto["runeSlotId"]


class RunePageData(CoreData):
    """This object contains rune page information."""
    _dto_type = RunePageDto
    _renamed = {"runes": "slots", "summoner_id": "summonerId"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def summoner_id(self) -> int:
        return self._dto["summonerId"]

    @property
    def current(self) -> bool:
        """Indicates if the rune page is the current rune page."""
        return self._dto["current"]

    @property
    def runes(self) -> List[RuneSlotData]:
        """Collection of rune slots associated with the rune page."""
        return [RuneSlotData.from_dto(slot) for slot in self._dto["slots"]]

    @property
    def name(self) -> str:
        """Rune page name."""
        return self._dto["name"]

    @property
    def id(self) -> int:
        """Rune page ID."""
        return self._dto["id"]


##############
# Core Types #
##############


class RunePages(CassiopeiaGhostList):
    _data_types = {RunePagesData}

    def __init__(self, *args, summoner: Union[Summoner, int, str], region: Union[Region, str] = None):
        super().__init__(*args, region=region)
        if isinstance(summoner, str):
            summoner = Summoner(name=summoner)
        elif isinstance(summoner, int):
            summoner = Summoner(id=summoner)
        self.__summoner = summoner

    def __get_query__(self):
        return {"summoner.id": self.__summoner.id, "region": self.region, "platform": self.platform}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        self.clear()
        from ..transformers.runes import RunesTransformer
        SearchableList.__init__(self, [RunesTransformer.rune_page_data_to_core(None, i) for i in data])
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[RunePagesData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform



@searchable({str: ["name", "runes", "region", "platform", "locale"], int: ["id"], bool: ["current"], StaticdataRune: ["runes"], Region: ["region"], Platform: ["platform"]})
class RunePage(CassiopeiaGhost):
    _data_types = {RunePageData}
    _load_types = {RunePageData: RunePages}

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[RunePageData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this rune page."""
        return self.region.platform

    @property
    def locale(self) -> str:
        """The locale for this rune page."""
        return self._data[RunePageData].locale

    @lazy_property
    def summoner(self) -> "Summoner":
        from .summoner import Summoner
        return Summoner(id=self._data[RunePageData].summoner_id)

    @CassiopeiaGhost.property(RunePageData)
    @ghost_load_on(KeyError)
    def current(self) -> bool:
        return self._data[RunePageData].current

    @CassiopeiaGhost.property(RunePageData)
    @ghost_load_on(KeyError)
    @lazy
    def runes(self) -> Mapping[StaticdataRune, int]:
        counter = Counter(slot.id for slot in self._data[RunePageData].runes)
        return SearchableDictionary({StaticdataRune(id=id, region=self.region.value): count for id, count in counter.items()})

    @CassiopeiaGhost.property(RunePageData)
    @ghost_load_on(KeyError)
    def name(self) -> str:
        return self._data[RunePageData].name

    @CassiopeiaGhost.property(RunePageData)
    @ghost_load_on(KeyError)
    def id(self) -> int:
        return self._data[RunePageData].id



from typing import Dict, Union

from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy_property
from merakicommons.container import searchable

from ...configuration import settings
from ...data import Region, Platform
from ..common import CoreData, CassiopeiaGhost
from ...dto.staticdata import realm as dto


##############
# Data Types #
##############


class RealmData(CoreData):
    _dto_type = dto.RealmDto
    _renamed = {"legacy_mode": "lg", "latest_data_dragon": "dd", "language": "l", "latest_versions": "n", "max_profile_icon_id": "profileiconmax", "store": "store", "version": "v", "cdn": "cdn", "css_version": "css"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def legacy_mode(self) -> str:
        """Legacy script mode for IE6 or older."""
        return self._dto["lg"]

    @property
    def latest_data_dragon(self) -> str:
        """Latest changed version of Dragon Magic."""
        return self._dto["dd"]

    @property
    def language(self) -> str:
        """Default language for this realm."""
        return self._dto["l"]

    @property
    def latest_versions(self) -> Dict[str, str]:
        """Latest changed version for each data type listed."""
        return self._dto["n"]

    @property
    def max_profile_icon_id(self) -> int:
        """Special behavior number identifying the largest profile icon ID that can be used under 500. Any profile icon that is requested between this number and 500 should be mapped to 0."""
        return self._dto["profileiconmax"]

    @property
    def store(self) -> str:
        """Additional API data drawn from other sources that may be related to Data Dragon functionality."""
        return self._dto["store"]

    @property
    def version(self) -> str:
        """Current version of this file for this realm."""
        return self._dto["v"]

    @property
    def cdn(self) -> str:
        """The base CDN URL."""
        return self._dto["cdn"]

    @property
    def css_version(self) -> str:
        """Latest changed version of Dragon Magics CSS file."""
        return self._dto["css"]


##############
# Core Types #
##############


@searchable({})
class Realms(CassiopeiaGhost):
    _data_types = {RealmData}

    def __init__(self, region: Union[Region, str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        kwargs = {"region": region}
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform}

    @lazy_property
    def region(self) -> Region:
        """The region for this realm."""
        return Region(self._data[RealmData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this realm."""
        return self.region.platform

    @lazy_property
    def locale(self) -> Platform:
        """The locale for this realm."""
        return self._data[RealmData].locale

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on(KeyError)
    def version(self) -> str:
        return self._data[RealmData].version

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on(KeyError)
    def language(self) -> str:
        """Default language for this realm."""
        return self._data[RealmData].language

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on(KeyError)
    def latest_versions(self) -> Dict[str, str]:
        """Latest changed version for each data type listed."""
        return self._data[RealmData].latest_versions

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on(KeyError)
    def legacy_mode(self) -> str:
        return self._data[RealmData].legacy_mode

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on(KeyError)
    def latest_data_dragon(self) -> str:
        return self._data[RealmData].latest_data_dragon

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on(KeyError)
    def language(self) -> str:
        return self._data[RealmData].language

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on(KeyError)
    def max_profile_icon_id(self) -> int:
        return self._data[RealmData].max_profile_icon_id

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on(KeyError)
    def store(self) -> str:
        return self._data[RealmData].store

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on(KeyError)
    def cdn(self) -> str:
        return self._data[RealmData].cdn

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on(KeyError)
    def css_version(self) -> str:
        return self._data[RealmData].css_version



from typing import Type, TypeVar, MutableMapping, Any, Iterable

from datapipelines import DataSource, PipelineContext, Query, NotFoundError
from .common import RiotAPIService, APINotFoundError
from ...data import Platform, Region
from ...dto.staticdata.version import VersionListDto
from ...dto.spectator import CurrentGameInfoDto, FeaturedGamesDto

T = TypeVar("T")


def _get_default_version(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    pipeline = context[PipelineContext.Keys.PIPELINE]
    versions = pipeline.get(VersionListDto, {"platform": query["platform"]})
    return versions["versions"][0]


def _get_default_locale(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    return query["platform"].default_locale


class SpectatorAPI(RiotAPIService):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    ################
    # Current Game #
    ################

    _validate_get_current_game_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.id").as_(int)

    @get.register(CurrentGameInfoDto)
    def get_current_game(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> CurrentGameInfoDto:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        SpectatorAPI._validate_get_current_game_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/spectator/v3/active-games/by-summoner/{id}".format(platform=query["platform"].value.lower(), id=query["summoner.id"])
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "spectator/active-games/by-summoner"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        return CurrentGameInfoDto(data)

    #################
    # Featured Game #
    #################

    _validate_get_featured_game_query = Query. \
        has("platform").as_(Platform)

    @get.register(FeaturedGamesDto)
    def get_featured_games(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> FeaturedGamesDto:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        SpectatorAPI._validate_get_featured_game_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/spectator/v3/featured-games".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "featured-games"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        for game in data["gameList"]:
            game["region"] = data["region"]
        return FeaturedGamesDto(data)



from typing import Type, TypeVar, MutableMapping, Any, Iterable, Generator

from datapipelines import DataSource, PipelineContext, Query, NotFoundError
from .common import RiotAPIService, APINotFoundError
from ...data import Platform, Region
from ...dto.staticdata.version import VersionListDto
from ...dto.status import ShardStatusDto

T = TypeVar("T")


def _get_default_version(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    pipeline = context[PipelineContext.Keys.PIPELINE]
    versions = pipeline.get(VersionListDto, {"platform": query["platform"]})
    return versions["versions"][0]


def _get_default_locale(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    return query["platform"].default_locale


class StatusAPI(RiotAPIService):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    ##########
    # Status #
    ##########

    _validate_get_status_query = Query. \
        has("platform").as_(Platform)

    @get.register(ShardStatusDto)
    def get_status(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ShardStatusDto:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        StatusAPI._validate_get_status_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/status/v3/shard-data".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "status"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        return ShardStatusDto(data)

    _validate_get_many_status_query = Query. \
        has("platforms").as_(Iterable)

    @get_many.register(ShardStatusDto)
    def get_many_status(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ShardStatusDto, None, None]:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        StatusAPI._validate_get_many_status_query(query, context)

        def generator():
            for platform in query["platforms"]:
                platform = Platform(platform.upper())
                url = "https://{platform}.api.riotgames.com/lol/status/v3/shard-data".format(platform=platform.value.lower())
                try:
                    data = self._get(url, {}, self._get_rate_limiter(query["platform"], "status"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["region"] = platform.region.value
                data["locale"] = query["locale"] if "locale" in query else platform.default_locale
                data["includedData"] = query["includedData"]
                yield ShardStatusDto(data)

        return generator()



import cassiopeia
import pytest

from datapipelines.common import NotFoundError

from .constants import SUMMONER_NAME, UNKNOWN_SUMMONER_NAME


def test_unknown_summoner():
    assert not cassiopeia.get_summoner(name=UNKNOWN_SUMMONER_NAME).exists
    assert not cassiopeia.get_summoner(id=9999999999999).exists


def test_equal_result_with_summoner_id():
    from_name = cassiopeia.get_summoner(name=SUMMONER_NAME)
    from_id = cassiopeia.get_summoner(id=from_name.id)

    assert from_name.id == from_id.id



import cassiopeia
import pytest

from merakicommons.container import SearchableList
from datapipelines.common import NotFoundError

from .constants import SUMMONER_NAME, UNKNOWN_SUMMONER_NAME


def test_matches_return_type():
    match_history = cassiopeia.get_match_history(SUMMONER_NAME)

    assert isinstance(match_history, SearchableList)
    assert all(isinstance(m, cassiopeia.Match) for m in match_history)


def test_matches_raises_with_unknown_summoner():
    with pytest.raises(NotFoundError):
        match_history = cassiopeia.get_match_history(UNKNOWN_SUMMONER_NAME)
        match_history[0]


def test_match_correct_return():
    match_history = cassiopeia.get_match_history(SUMMONER_NAME)
    first_match = match_history[0]
    first_match_data = first_match._data[cassiopeia.core.match.MatchData]._dto

    match_from_id = cassiopeia.get_match(first_match.id)
    match_from_id_data = match_from_id._data[cassiopeia.core.match.MatchData]._dto

    assert isinstance(match_from_id, cassiopeia.Match)
    #assert first_match_data == match_from_id_data
    assert first_match.id == match_from_id.id



from typing import List, Union, Set

from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable, SearchableList

from ...configuration import settings
from ...data import Resource, Region, Platform, GameMode
from ..common import CoreData, DataObjectList, CassiopeiaObject, CassiopeiaGhost, CassiopeiaGhostList, get_latest_version
from .common import ImageData, Image, Sprite
from ...dto.staticdata import summonerspell as dto


##############
# Data Types #
##############


class SummonerSpellListData(DataObjectList):
    _dto_type = dto.SummonerSpellListDto
    _renamed = {"included_data": "includedData"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]

    @property
    def included_data(self) -> Set[str]:
        return self._dto["includedData"]


class SpellVarsData(CoreData):
    _renamed = {"ranks_with": "ranksWith", "dynamic": "dyn", "coefficients": "coeff"}

    @property  # This doesn't get returned by the API
    def ranks_with(self) -> str:
        return self._dto["ranksWith"]

    @property  # This doesn't get returned by the API
    def dynamic(self) -> str:
        return self._dto["dyn"]

    @property
    def link(self) -> str:
        return self._dto["link"]

    @property
    def coefficients(self) -> List[float]:
        return self._dto["coeff"]

    @property
    def key(self) -> str:
        return self._dto["key"]


class LevelTipData(CoreData):
    _renamed = {"effects": "effect", "keywords": "label"}

    @property
    def effects(self) -> List[str]:
        return self._dto["effect"]

    @property
    def keywords(self) -> List[str]:
        return self._dto["label"]


class SummonerSpellData(CoreData):
    _dto_type = dto.SummonerSpellDto
    _renamed = {"variables": "vars", "sanitized_description": "sanitizedDescription", "sanitized_tooltip": "sanitizedTooltip", "max_rank": "maxrank", "cooldowns": "cooldown", "costs": "cost", "alternative_images": "altimages", "effects": "effect", "resource": "costType", "included_data": "includedData"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]

    @property
    def included_data(self) -> str:
        return self._dto["includedData"]

    @property
    def modes(self) -> List[str]:
        return self._dto["modes"]

    @property
    def variables(self) -> List[SpellVarsData]:
        return [SpellVarsData.from_dto(v) for v in self._dto["vars"]]

    @property
    def resource(self) -> str:
        return self._dto["costType"]

    @property
    def image(self) -> ImageData:
        return ImageData.from_dto(self._dto["image"])

    @property
    def sanitized_description(self) -> str:
        return self._dto["sanitizedDescription"]

    @property
    def sanitized_tooltip(self) -> str:
        return self._dto["sanitizedTooltip"]

    @property
    def effects(self) -> List[List[float]]:
        return self._dto["effect"]

    @property
    def tooltip(self) -> str:
        return self._dto["tooltip"]

    @property
    def max_rank(self) -> int:
        return self._dto["maxrank"]

    @property
    def range(self) -> List[Union[int, str]]:
        return self._dto["range"]

    @property
    def cooldowns(self) -> List[float]:
        return self._dto["cooldown"]

    @property
    def costs(self) -> List[int]:
        return self._dto["cost"]

    @property
    def key(self) -> str:
        return self._dto["key"]

    @property
    def description(self) -> str:
        return self._dto["description"]

    @property
    def alternative_images(self) -> List[ImageData]:
        return [ImageData.from_dto(alt) for alt in self._dto["altimages"]]

    @property
    def id(self) -> int:
        return self._dto["id"]

    @property
    def name(self) -> str:
        return self._dto["name"]


##############
# Core Types #
##############


class SummonerSpells(CassiopeiaGhostList):
    _data_types = {SummonerSpellListData}

    def __init__(self, *args, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        if included_data is None:
            included_data = {"all"}
        if locale is None:
            locale = region.default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if version:
            kwargs["version"] = version
        super().__init__(*args, **kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale, "includedData": self.included_data}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        self.clear()
        from ...transformers.staticdata import StaticDataTransformer
        SearchableList.__init__(self, [StaticDataTransformer.summoner_spell_data_to_core(None, ss) for ss in data])
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[SummonerSpellListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def version(self) -> str:
        try:
            return self._data[SummonerSpellListData].version
        except KeyError:
            version = get_latest_version(region=self.region, endpoint="summoner")
            self(version=version)
            return self._data[SummonerSpellListData].version

    @property
    def locale(self) -> str:
        """The locale for this champion."""
        return self._data[SummonerSpellListData].locale

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additonal information for this champion when it's loaded."""
        return self._data[SummonerSpellListData].included_data


@searchable({str: ["key"]})
class SpellVars(CassiopeiaObject):
    _data_types = {SpellVarsData}

    @property
    def ranks_with(self) -> str:
        """Well, we don't know what this one is. let us know if you figure it out."""
        return self._data[SpellVarsData].ranks_with

    @property
    def dynamic(self) -> str:
        """Well, we don't know what this one is. let us know if you figure it out."""
        return self._data[SpellVarsData].dynamic

    @property
    def link(self) -> str:
        """Stat this spell scales from."""
        return self._data[SpellVarsData].link

    @lazy_property
    def coefficients(self) -> List[float]:
        """The scaling coefficients for this spell."""
        return SearchableList(self._data[SpellVarsData].coefficients)

    @property
    def key(self) -> str:
        """Well, we don't know what this one is. let us know if you figure it out."""
        return self._data[SpellVarsData].key


@searchable({str: ["name", "key", "keywords", "resource"], Resource: ["resource"]})
class SummonerSpell(CassiopeiaGhost):
    _data_types = {SummonerSpellData}
    _load_types = {SummonerSpellData: SummonerSpellListData}

    def __init__(self, *, id: int = None, name: str = None, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        if included_data is None:
            included_data = {"all"}
        if locale is None:
            locale = region.default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if id is not None:
            kwargs["id"] = id
        if name is not None:
            kwargs["name"] = name
        if version is not None:
            kwargs["version"] = version
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale, "includedData": self.included_data}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        def find_matching_attribute(datalist, attrname, attrvalue):
            for item in datalist:
                if getattr(item, attrname, None) == attrvalue:
                    return item

        # The `data` is a dict of summoner spell data instances
        if "name" in self._data[SummonerSpellData]._dto:
            find = "name", self.name
        elif "id" in self._data[SummonerSpellData]._dto:
            find = "id", self.id
        else:
            raise RuntimeError("Expected fields not present after loading.")
        data = find_matching_attribute(data, *find)

        super().__load_hook__(load_group, data)

    # What do we do about params like this that can exist in both data objects?
    # They will be set on both data objects always, so we can choose either one to return.
    @lazy_property
    def region(self) -> Region:
        """The region for this summoner spell."""
        return Region(self._data[SummonerSpellData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this summoner spell."""
        return self.region.platform

    @property
    def version(self) -> str:
        """The version for this summoner spell."""
        try:
            return self._data[SummonerSpellData].version
        except KeyError:
            version = get_latest_version(region=self.region, endpoint="summoner")
            self(version=version)
            return self._data[SummonerSpellData].version

    @property
    def locale(self) -> str:
        """The locale for this summoner spell."""
        return self._data[SummonerSpellData].locale

    @property
    def included_data(self) -> Set[str]:
        """The data to included in the query for this summoner spell."""
        return self._data[SummonerSpellData].included_data

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    @lazy
    def modes(self) -> List[GameMode]:
        return SearchableList([GameMode(mode) for mode in self._data[SummonerSpellData].modes])

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    @lazy
    def variables(self) -> List[SpellVars]:
        """Contains spell data."""
        return SearchableList(SpellVars(v) for v in self._data[SummonerSpellData].variables)

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    @lazy
    def resource(self) -> Resource:
        """The resource consumed when using this spell."""
        return Resource(self._data[SummonerSpellData].resource)

    @lazy_property
    def sprite(self) -> Sprite:
        return self.image.sprite_info

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    def sanitized_description(self) -> str:
        """The spell's sanitized description."""
        return self._data[SummonerSpellData].sanitized_description

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    def sanitized_tooltip(self) -> str:
        """The spell's sanitized tooltip."""
        return self._data[SummonerSpellData].sanitized_tooltip

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    @lazy
    def effects(self) -> List[List[float]]:
        """The level-by-level replacements for {{ e# }} tags in other values."""
        return SearchableList(self._data[SummonerSpellData].effects)

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    def tooltip(self) -> str:
        """The spell's tooltip."""
        return self._data[SummonerSpellData].tooltip

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    def max_rank(self) -> int:
        """The maximum rank this spell can attain."""
        return self._data[SummonerSpellData].max_rank

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    @lazy
    def range(self) -> List[Union[int, str]]:
        """The maximum range of this spell. `self` if it has no range."""
        return SearchableList(self._data[SummonerSpellData].range)

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    @lazy
    def cooldowns(self) -> List[float]:
        """The cooldowns of this spell (per level)."""
        return SearchableList(self._data[SummonerSpellData].cooldowns)

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    @lazy
    def costs(self) -> List[int]:
        """The resource costs of this spell (per level)."""
        return SearchableList(self._data[SummonerSpellData].costs)

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    def key(self) -> str:
        """The spell's key."""
        return self._data[SummonerSpellData].key

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    def description(self) -> str:
        """The spell's description."""
        return self._data[SummonerSpellData].description

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    @lazy
    def alternative_images(self) -> List[Image]:
        """The alternative images for this spell. These won't exist after patch NN, when Riot standardized all images."""
        return SearchableList(Image(alt) for alt in self._data[SummonerSpellData].alternative_images)

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    def id(self) -> int:
        """The spell's id."""
        return self._data[SummonerSpellData].id

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    def name(self) -> str:
        """The spell's name."""
        return self._data[SummonerSpellData].name

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on(KeyError)
    @lazy
    def image(self) -> Image:
        image = Image(self._data[SummonerSpellData].image)
        image(version=self.version)
        return image

    @lazy_property
    def sprite(self) -> Sprite:
        return self.image.sprite_info



from typing import List, Dict, Union
import datetime

from datapipelines import NotFoundError
from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable, SearchableList, SearchableDictionary

from ..configuration import settings
from ..data import Region, Platform, GameMode, GameType, Queue, Map
from .common import CoreData, DataObjectList, CassiopeiaObject, CassiopeiaGhost, CassiopeiaGhostList
from ..dto import spectator as dto
from .staticdata.profileicon import ProfileIcon
from .staticdata.champion import Champion
from .staticdata.rune import Rune
from .staticdata.mastery import Mastery
from .staticdata.summonerspell import SummonerSpell
from .summoner import Summoner


##############
# Data Types #
##############


class FeaturedGamesData(DataObjectList):
    _dto_type = dto.FeaturedGamesDto
    _renamed = {"client_refresh_interval": "clientRefreshInterval"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def client_refresh_interval(self) -> str:
        return self._dto["clientRefreshInterval"]

    @property
    def summoner_id(self) -> str:
        return self._dto["summonerId"]


class RuneData(CoreData):
    _renamed = {"id": "runeId"}

    @property
    def id(self) -> int:
        """The ID of the rune"""
        return self._dto["runeId"]

    @property
    def count(self) -> int:
        """The count of this rune used by the participant"""
        return self._dto["count"]


class MasteryData(CoreData):
    _renamed = {"id": "masteryId", "points": "rank"}

    @property
    def id(self) -> int:
        """The ID of the mastery"""
        return self._dto["masteryId"]

    @property
    def points(self) -> int:
        """The number of points put into this mastery by the user"""
        return self._dto["rank"]


class CurrentGameParticipantData(CoreData):
    _renamed = {"profile_icon_id": "profileIconId", "champion_id": "championId", "summoner_name": "summonerName", "summoner_id": "summonerId", "bot": "bot", "team_id": "teamId", "summoner_spell_f": "spell1Id", "summoner_spell_d": "spell2Id"}

    @property
    def profile_icon_id(self) -> int:
        """The ID of the profile icon used by this participant"""
        return self._dto["profileIconId"]

    @property
    def champion_id(self) -> int:
        """The ID of the champion played by this participant"""
        return self._dto["championId"]

    @property
    def summoner_id(self) -> int:
        """The summoner ID of this participant"""
        return self._dto["summonerId"]

    @property
    def summoner_name(self) -> str:
        """The summoner name of this participant"""
        return self._dto["summonerName"]

    @property
    def runes(self) -> List[RuneData]:
        """The runes used by this participant"""
        return self._dto["runes"]

    @property
    def masteries(self) -> List[MasteryData]:
        """The masteries used by this participant"""
        return self._dto["masteries"]

    @property
    def is_bot(self) -> int:
        """Flag indicating whether or not this participant is a bot"""
        return self._dto["bot"]

    @property
    def team_id(self) -> int:
        """The team ID of this participant, indicating the participants team"""
        return self._dto["teamId"]

    @property
    def summoner_spell_d(self) -> int:
        """The ID of the first summoner spell used by this participant"""
        return self._dto["spell1Id"]

    @property
    def summoner_spell_f(self) -> int:
        """The ID of the second summoner spell used by this participant"""
        return self._dto["spell2Id"]


class TeamData(CoreData):
    _renamed = {}

    @property
    def participants(self) -> List[CurrentGameParticipantData]:
        return [CurrentGameParticipantData.from_dto(p) for p in self._dto["participants"]]

    @property
    def bans(self) -> Dict[int, int]:
        return {b["pickTurn"]: b["championId"] for b in self._dto["bans"]}


class CurrentGameInfoData(CoreData):
    _dto_type = dto.CurrentGameInfoDto
    _renamed = {"platform": "platformId", "observer_key": "observers", "creation": "gameStartTime", "duration": "gameLength", "mode": "gameMode", "map": "mapId", "type": "gameType", "queue": "gameQueueConfigId"}

    @property
    def id(self) -> int:
        return self._dto["gameId"]

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def teams(self) -> List[TeamData]:
        blue_team = {"participants": [], "bans": []}
        red_team = {"participants": [], "bans": []}
        for p in self._dto["participants"]:
            if p["teamId"] == 100:
                blue_team["participants"].append(p)
            elif p["teamId"] == 200:
                red_team["participants"].append(p)
        for b in self._dto["bannedChampions"]:
            if b["teamId"] == 100:
                blue_team["bans"].append(b)
            elif b["teamId"] == 200:
                red_team["bans"].append(b)
        return [TeamData.from_dto(blue_team), TeamData.from_dto(red_team)]

    @property
    def mode(self) -> str:
        return self._dto["gameMode"]

    @property
    def map(self) -> int:
        return self._dto["mapId"]

    @property
    def type(self) -> str:
        return self._dto["gameType"]

    @property
    def queue(self) -> int:
        return self._dto["gameQueueConfigId"]

    @property
    def duration(self) -> int:
        return self._dto["gameLength"]

    @property
    def creation(self) -> int:
        return self._dto["gameStartTime"]

    @property
    def observer_key(self) -> str:
        return self._dto["observers"]["encryptionKey"]


##############
# Core Types #
##############


class FeaturedMatches(CassiopeiaGhostList):
    _data_types = {FeaturedGamesData}

    def __get_query__(self):
        query = {"platform": self.platform}
        return query

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        self.clear()
        from ..transformers.spectator import SpectatorTransformer
        SearchableList.__init__(self, [SpectatorTransformer.current_game_data_to_core(None, i) for i in data])
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[FeaturedGamesData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def client_refresh_interval(self) -> int:
        return self._data[FeaturedGamesData].client_refresh_interval


@searchable({})
class Participant(CassiopeiaObject):
    _data_types = {CurrentGameParticipantData}

    @property
    def champion(self) -> Champion:
        return Champion(id=self._data[CurrentGameParticipantData].champion_id)

    @property
    def summoner(self) -> Summoner:
        ProfileIcon(id=self._data[CurrentGameParticipantData].profile_icon_id)
        return Summoner(id=self._data[CurrentGameParticipantData].summoner_id, name=self._data[CurrentGameParticipantData].summoner_name)

    @property
    def runes(self) -> Dict[Rune, int]:
        return SearchableDictionary({Rune(id=rune.id): rune.count for rune in self._data[CurrentGameParticipantData].runes})

    @property
    def masteries(self) -> Dict[Mastery, int]:
        return SearchableDictionary({Mastery(id=mastery.id): mastery.points for mastery in self._data[CurrentGameParticipantData].masteries})

    @property
    def is_bot(self) -> bool:
        return self._data[CurrentGameParticipantData].is_bot

    @property
    def team(self) -> "Team":
        raise NotImplemented  # TODO

    @property
    def summoner_spell_d(self) -> SummonerSpell:
        return SummonerSpell(id=self._data[CurrentGameParticipantData].summoner_spell_d)

    @property
    def summoner_spell_f(self) -> SummonerSpell:
        return SummonerSpell(id=self._data[CurrentGameParticipantData].summoner_spell_f)


@searchable({})
class Team(CassiopeiaObject):
    _data_types = {TeamData}

    @lazy_property
    def participants(self) -> List[Participant]:
        return SearchableList([Participant(p) for p in self._data[TeamData].participants])

    @lazy_property
    def bans(self) -> Dict[int, Champion]:
        return {pick: Champion(id=champion_id) for pick, champion_id in self._data[TeamData].bans.items()}


@searchable({})
class CurrentMatch(CassiopeiaGhost):
    _data_types = {CurrentGameInfoData}

    def __init__(self, *, summoner: Union[Summoner, int, str] = None, region: Union[Region, str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        kwargs = {"region": region, "id": id}

        if summoner is not None:
            if isinstance(summoner, str):
                summoner = Summoner(name=summoner)
            elif isinstance(summoner, int):
                summoner = Summoner(id=summoner)
            self.__summoner = summoner
        super().__init__(**kwargs)

    @classmethod
    def from_data(cls, data: CoreData, summoner: Union[Summoner, int, str]):
        self = super().from_data(data)
        if isinstance(summoner, str):
            summoner = Summoner(name=summoner)
        elif isinstance(summoner, int):
            summoner = Summoner(id=summoner)
        self.__summoner = summoner
        return self


    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "summoner.id": self.__summoner.id}

    @property
    def exists(self):
        try:
            if not self._Ghost__all_loaded:
                self.__load__()
            self.creation  # Make sure we can access this attribute
            return True
        except (AttributeError, NotFoundError):
            return False

    @lazy_property
    def region(self) -> Region:
        """The region for this summoner."""
        return Region(self._data[CurrentGameInfoData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this summoner."""
        return self.region.platform

    @CassiopeiaGhost.property(CurrentGameInfoData)
    @ghost_load_on(KeyError)
    def id(self) -> int:
        return self._data[CurrentGameInfoData].id

    @CassiopeiaGhost.property(CurrentGameInfoData)
    @ghost_load_on(KeyError)
    @lazy
    def teams(self) -> List[Team]:
        return SearchableList([Team(team) for team in self._data[CurrentGameInfoData].teams])

    @CassiopeiaGhost.property(CurrentGameInfoData)
    @ghost_load_on(KeyError)
    def blue_team(self) -> Team:
        return self._data[CurrentGameInfoData].teams[0]

    @CassiopeiaGhost.property(CurrentGameInfoData)
    @ghost_load_on(KeyError)
    def red_team(self) -> Team:
        return self._data[CurrentGameInfoData].teams[1]

    @CassiopeiaGhost.property(CurrentGameInfoData)
    @ghost_load_on(KeyError)
    def mode(self) -> GameMode:
        return GameMode(self._data[CurrentGameInfoData].mode)

    @CassiopeiaGhost.property(CurrentGameInfoData)
    @ghost_load_on(KeyError)
    def map(self) -> Map:
        return Map(self._data[CurrentGameInfoData].map)

    @CassiopeiaGhost.property(CurrentGameInfoData)
    @ghost_load_on(KeyError)
    def type(self) -> GameType:
        return GameType(self._data[CurrentGameInfoData].type)

    @CassiopeiaGhost.property(CurrentGameInfoData)
    @ghost_load_on(KeyError)
    def queue(self) -> Queue:
        return Queue(self._data[CurrentGameInfoData].queue)

    @CassiopeiaGhost.property(CurrentGameInfoData)
    @ghost_load_on(KeyError)
    def duration(self) -> datetime.timedelta:
        return datetime.timedelta(seconds=self._data[CurrentGameInfoData].duration)

    @CassiopeiaGhost.property(CurrentGameInfoData)
    @ghost_load_on(KeyError)
    def creation(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(self._data[CurrentGameInfoData].creation / 1000)

    @CassiopeiaGhost.property(CurrentGameInfoData)
    @ghost_load_on(KeyError)
    def observer_key(self) -> str:
        return self._data[CurrentGameInfoData].observer_key



from typing import Type, TypeVar, MutableMapping, Any, Iterable

from datapipelines import DataSource, PipelineContext, Query, NotFoundError
from .common import RiotAPIService, APINotFoundError
from ...data import Region, Platform
from ...dto.summoner import SummonerDto

T = TypeVar("T")


class SummonerAPI(RiotAPIService):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    _validate_get_summoner_query = Query. \
        has("id").as_(int). \
        or_("account.id").as_(int). \
        or_("name").as_(str).also. \
        has("platform").as_(Platform)

    @get.register(SummonerDto)
    def get_summoner(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> SummonerDto:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        SummonerAPI._validate_get_summoner_query(query, context)

        if "id" in query:
            url = "https://{platform}.api.riotgames.com/lol/summoner/v3/summoners/{summonerId}".format(platform=query["platform"].value.lower(), summonerId=query["id"])
            endpoint = "summoners/summonerId"
        elif "account.id" in query:
            url = "https://{platform}.api.riotgames.com/lol/summoner/v3/summoners/by-account/{accountId}".format(platform=query["platform"].value.lower(), accountId=query["account.id"])
            endpoint = "summoners/by-account/accountId"
        elif "name" in query:
            url = "https://{platform}.api.riotgames.com/lol/summoner/v3/summoners/by-name/{name}".format(platform=query["platform"].value.lower(), name=query["name"].replace(" ", ""))
            endpoint = "summoners/by-name/name"
        else:
            endpoint = ""

        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], endpoint))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        return SummonerDto(**data)



from typing import Type, TypeVar, MutableMapping, Any, Iterable, Generator

from datapipelines import DataSource, PipelineContext, Query, NotFoundError
from .common import RiotAPIService, APINotFoundError
from ...data import Platform
from ...dto.staticdata.champion import ChampionDto, ChampionListDto
from ...dto.staticdata.mastery import MasteryDto, MasteryListDto
from ...dto.staticdata.rune import RuneDto, RuneListDto
from ...dto.staticdata.item import ItemDto, ItemListDto
from ...dto.staticdata.summonerspell import SummonerSpellDto, SummonerSpellListDto
from ...dto.staticdata.version import VersionListDto
from ...dto.staticdata.map import MapListDto
from ...dto.staticdata.realm import RealmDto
from ...dto.staticdata.language import LanguagesDto, LanguageStringsDto
from ...dto.staticdata.profileicon import ProfileIconDataDto

T = TypeVar("T")


def _get_default_version(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    pipeline = context[PipelineContext.Keys.PIPELINE]
    realms = pipeline.get(RealmDto, {"platform": query["platform"]})
    return realms["version"]


def _get_default_locale(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    return query["platform"].default_locale


class StaticDataAPI(RiotAPIService):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    #############
    # Champions #
    #############

    _validate_get_champion_query = Query. \
        has("id").as_(int).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_default_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(ChampionDto)
    def get_champion(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionDto:
        StaticDataAPI._validate_get_champion_query(query, context)

        params = {
            "version": query["version"],
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"]))
        }

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/champions/{id}".format(platform=query["platform"].value.lower(), id=query["id"])
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/champion"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["version"] = query["version"]
        data["locale"] = query["locale"]
        data["includedData"] = query["includedData"]
        return ChampionDto(data)

    _validate_get_many_champion_query = Query. \
        has("ids").as_(Iterable).also. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get_many.register(ChampionDto)
    def get_many_champion(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ChampionDto, None, None]:
        StaticDataAPI._validate_get_many_champion_query(query, context)

        params = {
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"])),
            "dataById": True
        }

        if "version" in query:
            params["version"] = query["version"]

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/champions".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/champion"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        def generator():
            for id in query["ids"]:
                try:
                    champion = data["data"][str(id)]
                except KeyError as error:
                    raise NotFoundError("No champion exists with id \"{id}\"".format(id=id)) from error

                champion["region"] = query["platform"].region.value
                champion["version"] = data["version"]
                champion["locale"] = query["locale"]
                champion["includedData"] = query["includedData"]
                yield ChampionDto(champion)

        return generator()

    _validate_get_champion_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"}).also. \
        can_have("dataById").with_default(False)

    @get.register(ChampionListDto)
    def get_champion_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionListDto:
        StaticDataAPI._validate_get_champion_list_query(query, context)

        params = {
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"])),
            "dataById": query["dataById"]
        }

        if "version" in query:
            params["version"] = query["version"]

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/champions".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/champions"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["locale"] = query["locale"]
        data["includedData"] = query["includedData"]
        return ChampionListDto(data)

    _validate_get_many_champion_list_query = Query. \
        has("platforms").as_(Iterable).also. \
        can_have("version").as_(str).also. \
        can_have("locale").as_(str).also. \
        can_have("includedData").with_default({"all"}).also. \
        can_have("dataById").with_default(False)

    @get_many.register(ChampionListDto)
    def get_many_champion_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ChampionListDto, None, None]:
        StaticDataAPI._validate_get_many_champion_list_query(query, context)

        params = {
            "tags": ",".join(list(query["includedData"])),
            "dataById": query["dataById"]
        }

        if "version" in query:
            params["version"] = query["version"]

        if "locale" in query:
            params["locale"] = query["locale"]

        def generator():
            for platform in query["platforms"]:
                platform = Platform(platform.upper())
                url = "https://{platform}.api.riotgames.com/lol/static-data/v3/champions".format(platform=platform.value.lower())
                try:
                    data = self._get(url, params, self._get_rate_limiter(platform, "staticdata/champions"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["region"] = platform.region.value
                data["locale"] = query["locale"] if "locale" in query else platform.default_locale
                data["includedData"] = query["includedData"]
                yield ChampionListDto(data)

        return generator()

    ############
    # Versions #
    ############

    _validate_get_versions_query = Query. \
        has("platform").as_(Platform)

    @get.register(VersionListDto)
    def get_versions(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> VersionListDto:
        StaticDataAPI._validate_get_versions_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/versions".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "staticdata/versions"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        return VersionListDto({
            "region": query["platform"].region.value,
            "versions": data
        })

    #############
    # Masteries #
    #############

    _validate_get_mastery_query = Query. \
        has("id").as_(int).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_default_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(MasteryDto)
    def get_mastery(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MasteryDto:
        StaticDataAPI._validate_get_mastery_query(query, context)

        params = {
            "version": query["version"],
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"]))
        }

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/masteries/{id}".format(platform=query["platform"].value.lower(), id=query["id"])
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/mastery"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["version"] = query["version"]
        data["locale"] = query["locale"]
        data["includedData"] = query["includedData"]
        return MasteryDto(data)

    _validate_get_many_mastery_query = Query. \
        has("ids").as_(Iterable).also. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get_many.register(MasteryDto)
    def get_many_mastery(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[MasteryDto, None, None]:
        StaticDataAPI._validate_get_many_mastery_query(query, context)

        params = {
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"]))
        }

        if "version" in query:
            params["version"] = query["version"]

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/masteries".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/mastery"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        def generator():
            for id in query["ids"]:
                try:
                    mastery = data["data"][str(id)]
                except KeyError as error:
                    raise NotFoundError("No mastery exists with id \"{id}\"".format(id=id)) from error

                mastery["region"] = query["platform"].region.value
                mastery["version"] = data["version"]
                mastery["locale"] = query["locale"]
                mastery["includedData"] = query["includedData"]
                yield MasteryDto(mastery)

        return generator()

    _validate_get_mastery_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(MasteryListDto)
    def get_mastery_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MasteryListDto:
        StaticDataAPI._validate_get_mastery_list_query(query, context)

        params = {
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"]))
        }

        if "version" in query:
            params["version"] = query["version"]

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/masteries".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/masteries"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["locale"] = query["locale"]
        data["includedData"] = query["includedData"]
        return MasteryListDto(data)

    _validate_get_many_mastery_list_query = Query. \
        has("platforms").as_(Iterable).also. \
        can_have("version").as_(str).also. \
        can_have("locale").as_(str).also. \
        can_have("includedData").with_default({"all"})

    @get_many.register(MasteryListDto)
    def get_many_mastery_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[MasteryListDto, None, None]:
        StaticDataAPI._validate_get_many_mastery_list_query(query, context)

        params = {
            "tags": ",".join(list(query["includedData"]))
        }

        if "version" in query:
            params["version"] = query["version"]

        if "locale" in query:
            params["locale"] = query["locale"]

        def generator():
            for platform in query["platforms"]:
                platform = Platform(platform.upper())
                url = "https://{platform}.api.riotgames.com/lol/static-data/v3/masteries".format(platform=platform.value.lower())
                try:
                    data = self._get(url, params, self._get_rate_limiter(platform, "staticdata/masteries"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["region"] = platform.region.value
                data["locale"] = query["locale"] if "locale" in query else platform.default_locale
                data["includedData"] = query["includedData"]
                yield MasteryListDto(data)

        return generator()

    #########
    # Runes #
    #########

    _validate_get_rune_query = Query. \
        has("id").as_(int).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_default_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(RuneDto)
    def get_rune(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> RuneDto:
        StaticDataAPI._validate_get_rune_query(query, context)

        params = {
            "version": query["version"],
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"]))
        }

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/runes/{id}".format(platform=query["platform"].value.lower(), id=query["id"])
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/rune"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["version"] = query["version"]
        data["locale"] = query["locale"]
        data["includedData"] = query["includedData"]
        return RuneDto(data)

    _validate_get_many_rune_query = Query. \
        has("ids").as_(Iterable).also. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get_many.register(RuneDto)
    def get_many_rune(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[RuneDto, None, None]:
        StaticDataAPI._validate_get_many_rune_query(query, context)

        params = {
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"]))
        }

        if "version" in query:
            params["version"] = query["version"]

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/runes".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/rune"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        def generator():
            for id in query["ids"]:
                try:
                    rune = data["data"][str(id)]
                except KeyError as error:
                    raise NotFoundError("No rune exists with id \"{id}\"".format(id=id)) from error

                rune["region"] = query["platform"].region.value
                rune["version"] = data["version"]
                rune["locale"] = query["locale"]
                rune["includedData"] = query["includedData"]
                yield RuneDto(rune)

        return generator()

    _validate_get_rune_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(RuneListDto)
    def get_rune_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> RuneListDto:
        StaticDataAPI._validate_get_rune_list_query(query, context)

        params = {
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"]))
        }

        if "version" in query:
            params["version"] = query["version"]

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/runes".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/runes"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["locale"] = query["locale"]
        data["includedData"] = query["includedData"]
        return RuneListDto(data)

    _validate_get_many_rune_list_query = Query. \
        has("platforms").as_(Iterable).also. \
        can_have("version").as_(str).also. \
        can_have("locale").as_(str).also. \
        can_have("includedData").with_default({"all"})

    @get_many.register(RuneListDto)
    def get_many_rune_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[RuneListDto, None, None]:
        StaticDataAPI._validate_get_many_rune_list_query(query, context)

        params = {
            "tags": ",".join(list(query["includedData"]))
        }

        if "version" in query:
            params["version"] = query["version"]

        if "locale" in query:
            params["locale"] = query["locale"]

        def generator():
            for platform in query["platforms"]:
                platform = Platform(platform.upper())
                url = "https://{platform}.api.riotgames.com/lol/static-data/v3/runes".format(platform=platform.value.lower())
                try:
                    data = self._get(url, params, self._get_rate_limiter(platform, "staticdata/runes"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["region"] = platform.region.value
                data["locale"] = query["locale"] if "locale" in query else platform.default_locale
                data["includedData"] = query["includedData"]
                yield RuneListDto(data)

        return generator()

    #########
    # Items #
    #########

    _validate_get_item_query = Query. \
        has("id").as_(int).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_default_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(ItemDto)
    def get_item(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ItemDto:
        StaticDataAPI._validate_get_item_query(query, context)

        params = {
            "version": query["version"],
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"]))
        }

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/items/{id}".format(platform=query["platform"].value.lower(), id=query["id"])
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/item"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["version"] = query["version"]
        data["locale"] = query["locale"]
        data["includedData"] = query["includedData"]
        if data["id"] == 3632:  # This item doesn't have a name.
            data["name"] = ""
        if "tags" not in data:
            data["tags"] = []
        if "depth" not in data:
            data["depth"] = 1
        if "colloq" not in data:
            data["colloq"] = ""
        if "plaintext" not in data:
            data["plaintext"] = ""
        return ItemDto(data)

    _validate_get_many_item_query = Query. \
        has("ids").as_(Iterable).also. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get_many.register(ItemDto)
    def get_many_item(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ItemDto, None, None]:
        StaticDataAPI._validate_get_many_item_query(query, context)

        params = {
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"]))
        }

        if "version" in query:
            params["version"] = query["version"]

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/items".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/item"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        def generator():
            for id in query["ids"]:
                try:
                    item = data["data"][str(id)]
                except KeyError as error:
                    raise NotFoundError("No item exists with id \"{id}\"".format(id=id)) from error

                item["region"] = query["platform"].region.value
                item["version"] = data["version"]
                item["locale"] = query["locale"]
                item["includedData"] = query["includedData"]
                if item["id"] == 3632:  # This item doesn't have a name.
                    item["name"] = ""
                if "tags" not in item:
                    item["tags"] = []
                if "depth" not in item:
                    item["depth"] = 1
                if "colloq" not in item:
                    item["colloq"] = ""
                if "plaintext" not in item:
                    item["plaintext"] = ""
                yield ItemDto(item)

        return generator()

    _validate_get_item_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(ItemListDto)
    def get_item_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ItemListDto:
        StaticDataAPI._validate_get_item_list_query(query, context)

        params = {
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"]))
        }

        if "version" in query:
            params["version"] = query["version"]

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/items".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/items"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["locale"] = query["locale"]
        data["includedData"] = query["includedData"]
        for item in data["data"].values():
            if item["id"] == 3632:  # This item doesn't have a name.
                item["name"] = ""
            if "tags" not in item:
                item["tags"] = []
            if "depth" not in item:
                item["depth"] = 1
            if "colloq" not in item:
                item["colloq"] = ""
            if "plaintext" not in item:
                item["plaintext"] = ""
        return ItemListDto(data)

    _validate_get_many_item_list_query = Query. \
        has("platforms").as_(Iterable).also. \
        can_have("version").as_(str).also. \
        can_have("locale").as_(str).also. \
        can_have("includedData").with_default({"all"})

    @get_many.register(ItemListDto)
    def get_many_item_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ItemListDto, None, None]:
        StaticDataAPI._validate_get_many_item_list_query(query, context)

        params = {
            "tags": ",".join(list(query["includedData"]))
        }

        if "version" in query:
            params["version"] = query["version"]

        if "locale" in query:
            params["locale"] = query["locale"]

        def generator():
            for platform in query["platforms"]:
                platform = Platform(platform.upper())
                url = "https://{platform}.api.riotgames.com/lol/static-data/v3/items".format(platform=platform.value.lower())
                try:
                    data = self._get(url, params, self._get_rate_limiter(platform, "staticdata/items"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["region"] = platform.region.value
                data["locale"] = query["locale"] if "locale" in query else platform.default_locale
                data["includedData"] = query["includedData"]
                for item in data["data"].values():
                    if item["id"] == 3632:  # This item doesn't have a name.
                        item["name"] = ""
                    if "tags" not in item:
                        item["tags"] = []
                    if "depth" not in item:
                        item["depth"] = 1
                    if "colloq" not in item:
                        item["colloq"] = ""
                    if "plaintext" not in item:
                        item["plaintext"] = ""
                yield ItemListDto(data)

        return generator()

    ########
    # Maps #
    ########

    _validate_get_map_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(MapListDto)
    def get_map_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MapListDto:
        StaticDataAPI._validate_get_map_list_query(query, context)

        params = {
            "locale": query["locale"]
        }

        if "version" in query:
            params["version"] = query["version"]

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/maps".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/maps"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["locale"] = query["locale"]
        return MapListDto(data)

    ###################
    # Summoner Spells #
    ###################

    _validate_get_summoner_spell_query = Query. \
        has("id").as_(int).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_default_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(SummonerSpellDto)
    def get_summoner_spell(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> SummonerSpellDto:
        StaticDataAPI._validate_get_summoner_spell_query(query, context)

        params = {
            "version": query["version"],
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"]))
        }

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/summoner-spells/{id}".format(platform=query["platform"].value.lower(), id=query["id"])
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/summoner-spell"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["version"] = query["version"]
        data["locale"] = query["locale"]
        data["includedData"] = query["includedData"]
        return SummonerSpellDto(data)

    _validate_get_many_summoner_spell_query = Query. \
        has("ids").as_(Iterable).also. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get_many.register(SummonerSpellDto)
    def get_many_summoner_spell(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[SummonerSpellDto, None, None]:
        StaticDataAPI._validate_get_many_summoner_spell_query(query, context)

        params = {
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"]))
        }

        if "version" in query:
            params["version"] = query["version"]

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/summoner-spells".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/summoner-spell"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        def generator():
            for id in query["ids"]:
                try:
                    summoner_spell = data["data"][id]
                except KeyError as error:
                    raise NotFoundError("No summoner spell exists with id \"{id}\"".format(id=id)) from error

                summoner_spell["region"] = query["platform"].region.value
                summoner_spell["version"] = data["version"]
                summoner_spell["locale"] = query["locale"]
                summoner_spell["includedData"] = query["includedData"]
                yield SummonerSpellDto(summoner_spell)

        return generator()

    _validate_get_summoner_spell_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(SummonerSpellListDto)
    def get_summoner_spell_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> SummonerSpellListDto:
        StaticDataAPI._validate_get_summoner_spell_list_query(query, context)

        params = {
            "locale": query["locale"],
            "tags": ",".join(list(query["includedData"]))
        }

        if "version" in query:
            params["version"] = query["version"]

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/summoner-spells".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/summoner-spells"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["locale"] = query["locale"]
        data["includedData"] = query["includedData"]
        return SummonerSpellListDto(data)

    _validate_get_many_summoner_spell_list_query = Query. \
        has("platforms").as_(Iterable).also. \
        can_have("version").as_(str).also. \
        can_have("locale").as_(str).also. \
        can_have("includedData").with_default({"all"})

    @get_many.register(SummonerSpellListDto)
    def get_many_summoner_spell_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[SummonerSpellListDto, None, None]:
        StaticDataAPI._validate_get_many_summoner_spell_list_query(query, context)

        params = {
            "tags": ",".join(list(query["includedData"]))
        }

        if "version" in query:
            params["version"] = query["version"]

        if "locale" in query:
            params["locale"] = query["locale"]

        def generator():
            for platform in query["platforms"]:
                platform = Platform(platform.upper())
                url = "https://{platform}.api.riotgames.com/lol/static-data/v3/summoner-spells".format(platform=platform.value.lower())
                try:
                    data = self._get(url, params, self._get_rate_limiter(platform, "staticdata/summoner-spells"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["region"] = platform.region.value
                data["locale"] = query["locale"] if "locale" in query else platform.default_locale
                data["includedData"] = query["includedData"]
                yield SummonerSpellListDto(data)

        return generator()

    ##########
    # Realms #
    ##########

    _validate_get_realms_query = Query. \
        has("platform").as_(Platform)

    @get.register(RealmDto)
    def get_realms(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> RealmDto:
        StaticDataAPI._validate_get_realms_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/realms".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "staticdata/realms"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        return RealmDto(data)

    _validate_get_many_realms_query = Query. \
        has("platforms").as_(Iterable)

    @get_many.register(RealmDto)
    def get_many_realms(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[RealmDto, None, None]:
        StaticDataAPI._validate_get_many_realms_query(query, context)

        def generator():
            for platform in query["platforms"]:
                platform = Platform(platform.upper())
                url = "https://{platform}.api.riotgames.com/lol/static-data/v3/realms".format(platform=platform.value.lower())
                try:
                    data = self._get(url, {}, self._get_rate_limiter(platform, "staticdata/realms"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["region"] = platform.region.value
                yield RealmDto(data)

        return generator()

    ############
    # Language #
    ############

    _validate_get_languages_query = Query. \
        has("platform").as_(Platform)

    @get.register(LanguagesDto)
    def get_language(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> LanguagesDto:
        StaticDataAPI._validate_get_languages_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/languages".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "staticdata/language"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data = {"region": query["platform"].region.value, "languages": data}
        return LanguagesDto(data)

    _validate_get_many_languages_query = Query. \
        has("platforms").as_(Iterable)

    @get_many.register(LanguagesDto)
    def get_many_language(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[LanguagesDto, None, None]:
        StaticDataAPI._validate_get_many_languages_query(query, context)

        def generator():
            for platform in query["platforms"]:
                platform = Platform(platform.upper())
                url = "https://{platform}.api.riotgames.com/lol/static-data/v3/languages".format(platform=platform.value.lower())
                try:
                    data = self._get(url, {}, self._get_rate_limiter(platform, "staticdata/language"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["region"] = platform.region.value
                yield LanguagesDto(data)

        return generator()

    ####################
    # Language Strings #
    ####################

    _validate_get_language_strings_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str)

    @get.register(LanguageStringsDto)
    def get_language_strings(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> LanguageStringsDto:
        StaticDataAPI._validate_get_language_strings_query(query, context)

        params = {
            "locale": query["locale"]
        }

        if "version" in query:
            params["version"] = query["version"]

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/language-strings".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/language-strings"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["locale"] = query["locale"]
        return LanguageStringsDto(data)

    _validate_get_many_language_strings_query = Query. \
        has("platforms").as_(Iterable).also. \
        can_have("version").as_(str).also. \
        can_have("locale").as_(str)

    @get_many.register(LanguageStringsDto)
    def get_many_language_strings(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[LanguageStringsDto, None, None]:
        StaticDataAPI._validate_get_many_language_strings_query(query, context)

        params = {}

        if "version" in query:
            params["version"] = query["version"]

        if "locale" in query:
            params["locale"] = query["locale"]

        def generator():
            for platform in query["platforms"]:
                platform = Platform(platform.upper())
                url = "https://{platform}.api.riotgames.com/lol/static-data/v3/language-strings".format(platform=platform.value.lower())
                try:
                    data = self._get(url, params, self._get_rate_limiter(platform, "staticdata/language-strings"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["region"] = platform.region.value
                data["locale"] = query["locale"] if "locale" in query else platform.default_locale
                yield LanguageStringsDto(data)

        return generator()

    #################
    # Profile Icons #
    #################

    _validate_get_profile_icons_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str)

    @get.register(ProfileIconDataDto)
    def get_profile_icons(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ProfileIconDataDto:
        StaticDataAPI._validate_get_profile_icons_query(query, context)

        params = {
            "locale": query["locale"]
        }

        if "version" in query:
            params["version"] = query["version"]

        url = "https://{platform}.api.riotgames.com/lol/static-data/v3/profile-icons".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "staticdata/profile-icons"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["locale"] = query["locale"]
        return ProfileIconDataDto(data)

    _validate_get_many_profile_icons_query = Query. \
        has("platforms").as_(Iterable).also. \
        can_have("version").as_(str).also. \
        can_have("locale").as_(str)

    @get_many.register(ProfileIconDataDto)
    def get_many_profile_icons(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ProfileIconDataDto, None, None]:
        StaticDataAPI._validate_get_many_profile_icons_query(query, context)

        params = {}

        if "version" in query:
            params["version"] = query["version"]

        if "locale" in query:
            params["locale"] = query["locale"]

        def generator():
            for platform in query["platforms"]:
                platform = Platform(platform.upper())
                url = "https://{platform}.api.riotgames.com/lol/static-data/v3/profile-icons".format(platform=platform.value.lower())
                try:
                    data = self._get(url, params, self._get_rate_limiter(platform, "staticdata/profile-icons"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["region"] = platform.region.value
                data["locale"] = query["locale"] if "locale" in query else platform.default_locale
                yield ProfileIconDataDto(data)

        return generator()



from typing import List, Union

from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy_property, lazy
from merakicommons.container import searchable, SearchableList

from ..configuration import settings
from ..data import Region, Platform, Tier, Division, Queue
from .common import CoreData, DataObjectList, CassiopeiaObject, CassiopeiaGhost, CassiopeiaGhostList
from ..dto.league import LeaguesListDto, LeagueListDto, LeagueItemDto, MiniSeriesDto
from .summoner import Summoner


##############
# Data Types #
##############


class MiniSeriesData(CoreData):
    _dto_type = MiniSeriesDto
    _renamed = {}

    @property
    def wins(self) -> int:
        return self._dto["wins"]

    @property
    def losses(self) -> int:
        return self._dto["losses"]

    @property
    def target(self) -> int:
        return self._dto["target"]

    @property
    def progress(self) -> str:
        return self._dto["progress"]


class LeagueItemData(CoreData):
    _dto_type = LeagueItemDto
    _renamed = {"hot_streak": "hotStreak", "promos": "miniSeries", "summoner_id": "playerOrTeamId", "summoner_name": "playerOrTeamName", "league_points": "leaguePoints", "name": "leagueName", "queue": "queueType", "division": "rank", "fresh_blood": "freshBlood"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def tier(self) -> str:
        return self._dto["tier"]

    @property
    def queue(self) -> str:
        return self._dto["queue"]

    @property
    def name(self) -> str:
        return self._dto["leagueName"]

    @property
    def division(self) -> str:
        return self._dto["rank"]

    @property
    def hot_streak(self) -> bool:
        return self._dto["hotStreak"]

    @property
    def promos(self) -> MiniSeriesData:
        return MiniSeriesData.from_dto(self._dto["miniSeries"])

    @property
    def wins(self) -> int:
        return self._dto["wins"]

    @property
    def veteran(self) -> bool:
        return self._dto["veteran"]

    @property
    def losses(self) -> int:
        return self._dto["losses"]

    @property
    def summoner_id(self) -> int:
        return int(self._dto["playerOrTeamId"])

    @property
    def summoner_name(self) -> str:
        return self._dto["playerOrTeamName"]

    @property
    def inactive(self) -> bool:
        return self._dto["inactive"]

    @property
    def fresh_blood(self) -> bool:
        return self._dto["freshBlood"]

    @property
    def league_points(self) -> int:
        return self._dto["leaguePoints"]


class LeagueListData(CoreData):
    _dto_type = LeagueListDto
    _renamed = {}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def tier(self) -> str:
        return self._dto["tier"]

    @property
    def queue(self) -> str:
        return self._dto["queue"]

    @property
    def name(self) -> str:
        return self._dto["name"]

    @property
    def entries(self) -> List[LeagueItemData]:
        return [LeagueItemData.from_dto(entry) for entry in self._dto["entries"]]


class LeaguesListData(DataObjectList):
    _dto_type = LeaguesListDto
    _renamed = {"summoner_id": "summonerId"}

    @property
    def summoner_id(self) -> str:
        return self._dto["summonerId"]

    @property
    def region(self) -> str:
        return self._dto["region"]


class ChallengerLeagueListData(LeagueListData):
    pass


class MasterLeagueListData(LeagueListData):
    pass


##############
# Core Types #
##############


class MiniSeries(CassiopeiaObject):
    _data_types = {MiniSeriesData}
    # Technically wins, loses, and wins_required can all be calculated from progress, so we don't technically need to store those data.

    @property
    def wins(self) -> int:
        return self._data[MiniSeriesData].wins  # sum(self.progress)

    @property
    def loses(self) -> int:
        return self._data[MiniSeriesData].loses  # len(self._data[MiniSeriesData].progress) - sum(self.progress)

    @property
    def wins_required(self) -> int:
        """2 or 3 wins will be required for promotion."""
        return self._data[MiniSeriesData].target  # len(self._data[MiniSeriesData].progress)

    @lazy_property
    def progress(self) -> List[bool]:
        """A list of True/False for the number of games the played in the mini series indicating if the player won or lost."""
        return [True if p == "W" else False for p in self._data[MiniSeriesData].progress if p is not None]


@searchable({str: ["division", "name", "summoner"], bool: ["hot_streak", "veteran", "fresh_blood"], Division: ["division"], Summoner: ["summoner"]})
class LeagueEntry(CassiopeiaObject):
    _data_types = {LeagueItemData}

    def __init__(self, *, region: Union[Region, str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        kwargs = {"region": region}
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform}

    @lazy_property
    def region(self) -> Region:
        """The region for this champion."""
        return Region(self._data[LeagueItemData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this champion."""
        return self.region.platform

    @lazy_property
    def tier(self) -> Tier:
        return Tier(self._data[LeagueItemData].tier)

    @lazy_property
    def queue(self) -> Queue:
        return Queue(self._data[LeagueItemData].queue)

    @property
    def name(self) -> str:
        return self._data[LeagueItemData].name

    @lazy_property
    def division(self) -> Division:
        return Division(self._data[LeagueItemData].division)

    @property
    def hot_streak(self) -> bool:
        return self._data[LeagueItemData].hot_streak

    @lazy_property
    def promos(self) -> MiniSeries:
        return MiniSeries.from_data(self._data[LeagueItemData].promos)

    @property
    def wins(self) -> int:
        return self._data[LeagueItemData].wins

    @property
    def veteran(self) -> bool:
        return self._data[LeagueItemData].veteran

    @property
    def losses(self) -> int:
        return self._data[LeagueItemData].losses

    @lazy_property
    def summoner(self) -> Summoner:
        return Summoner(id=self._data[LeagueItemData].summoner_id, name=self._data[LeagueItemData].summoner_name)

    @property
    def fresh_blood(self) -> bool:
        return self._data[LeagueItemData].fresh_blood

    @property
    def league_points(self) -> int:
        return self._data[LeagueItemData].league_points


@searchable({str: ["tier", "queue", "name"], Queue: ["queue"], Tier: ["tier"]})
class League(CassiopeiaObject):
    _data_types = {LeagueListData}

    def __getitem__(self, item):
        return self.entries[item]

    def __len__(self):
        return len(self.entries)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[LeaguesListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @lazy_property
    def tier(self) -> Tier:
        return Tier(self._data[LeagueListData].tier)

    @lazy_property
    def queue(self) -> Queue:
        return Queue(self._data[LeagueListData].queue)

    @property
    def name(self) -> str:
        return self._data[LeagueListData].name

    @lazy_property
    def entries(self) -> List[LeagueEntry]:
        return SearchableList([LeagueEntry.from_data(entry) for entry in self._data[LeagueListData].entries])


class Leagues(CassiopeiaGhostList):
    _data_types = {LeaguesListData}

    def __init__(self, *args, summoner: Union[Summoner, int, str], region: Union[Region, str] = None):
        super().__init__(*args, region=region)
        if isinstance(summoner, str):
            summoner = Summoner(name=summoner)
        elif isinstance(summoner, int):
            summoner = Summoner(id=summoner)
        self.__summoner = summoner

    def __get_query__(self):
        return {"summoner.id": self.__summoner.id, "region": self.region, "platform": self.platform}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        self.clear()
        from ..transformers.leagues import LeagueTransformer
        SearchableList.__init__(self, [LeagueTransformer.league_data_to_core(None, i) for i in data])
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[LeaguesListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def fives(self):
        return self[Queue.ranked_solo]

    @property
    def flex(self):
        return self[Queue.flex]

    @property
    def threes(self):
        return self[Queue.ranked_threes]


class ChallengerLeague(League, CassiopeiaGhost):
    _data_types = {ChallengerLeagueListData}

    def __init__(self, *, queue: Union[Queue, str, int] = None, region: Union[Region, str] = None):
        kwargs = {"region": region}
        if isinstance(queue, int):
            kwargs["queue"] = Queue.from_id(queue)
        elif isinstance(queue, str):
            kwargs["queue"] = Queue(queue)
        elif isinstance(queue, Queue):
            kwargs["queue"] = queue
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "queue": self.queue}

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[ChallengerLeagueListData].region)

    @lazy_property
    def tier(self) -> Tier:
        return Tier.challenger

    @lazy_property
    def queue(self) -> Queue:
        return Queue(self._data[ChallengerLeagueListData].queue)

    @CassiopeiaGhost.property(ChallengerLeagueListData)
    @ghost_load_on(KeyError)
    def name(self) -> str:
        return self._data[ChallengerLeagueListData].name

    @CassiopeiaGhost.property(ChallengerLeagueListData)
    @ghost_load_on(KeyError)
    @lazy
    def entries(self) -> List[LeagueEntry]:
        return SearchableList([LeagueEntry.from_data(entry) for entry in self._data[ChallengerLeagueListData].entries])


class MasterLeague(League, CassiopeiaGhost):
    _data_types = {MasterLeagueListData}

    def __init__(self, *, queue: Union[Queue, str, int] = None, region: Union[Region, str] = None):
        kwargs = {"region": region}
        if isinstance(queue, int):
            kwargs["queue"] = Queue.from_id(queue)
        elif isinstance(queue, str):
            kwargs["queue"] = Queue(queue)
        elif isinstance(queue, Queue):
            kwargs["queue"] = queue
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "queue": self.queue}

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[MasterLeagueListData].region)

    @lazy_property
    def tier(self) -> Tier:
        return Tier.master

    @lazy_property
    def queue(self) -> Queue:
        return Queue(self._data[MasterLeagueListData].queue)

    @CassiopeiaGhost.property(MasterLeagueListData)
    @ghost_load_on(KeyError)
    def name(self) -> str:
        return self._data[MasterLeagueListData].name

    @CassiopeiaGhost.property(MasterLeagueListData)
    @ghost_load_on(KeyError)
    @lazy
    def entries(self) -> List[LeagueEntry]:
        return SearchableList([LeagueEntry.from_data(entry) for entry in self._data[MasterLeagueListData].entries])



from typing import List, Union
import datetime

from .data import PATCHES, Region, Queue
from .patches import Patch
from .core import Champion, Summoner, Account, ChampionMastery, Rune, Mastery, Item, RunePage, MasteryPage, Match, Map, SummonerSpell, Realms, ProfileIcon, LanguageStrings, CurrentMatch, ShardStatus, Versions, MatchHistory, Champions, ChampionMasteries, Masteries, Runes, Items, SummonerSpells, Maps, FeaturedMatches, Locales, ProfileIcons, MasteryPages, RunePages, ChallengerLeague, MasterLeague, Leagues


def get_leagues(summoner: Union[Summoner, int, str], region: Union[Region, str] = None) -> Leagues:
    return Leagues(summoner=summoner, region=region)


def get_master_league(queue: Union[Queue, int, str], region: Union[Region, str] = None) -> MasterLeague:
    return MasterLeague(queue=queue, region=region)


def get_challenger_league(queue: Union[Queue, int, str], region: Union[Region, str] = None) -> ChallengerLeague:
    return ChallengerLeague(queue=queue, region=region)


def get_match_history(summoner: Union[Summoner, int, str], region: Union[Region, str] = None) -> MatchHistory:
    return MatchHistory(summoner=summoner, region=region)


def get_match(id, region: Union[Region, str] = None) -> Match:
    return Match(id=id, region=region)


def get_featured_matches(region: Union[Region, str] = None) -> FeaturedMatches:
    return FeaturedMatches(region=region)


def get_current_match(summoner: Union[Summoner, int, str], region: Union[Region, str] = None) ->  CurrentMatch:
    return CurrentMatch(summoner=summoner, region=region)


def get_champion_masteries(summoner: Union[Summoner, int, str], region: Union[Region, str] = None) -> ChampionMasteries:
    return ChampionMasteries(summoner=summoner, region=region)


def get_champion_mastery(summoner: Union[Summoner, int, str], champion: Union[Champion, int, str], region: Union[Region, str] = None) -> ChampionMastery:
    return ChampionMastery(champion=champion, summoner=summoner, region=region)


def get_summoner(*, id: int = None, account: Union[Account, int] = None, name: str = None, region: Union[Region, str] = None) -> Summoner:
    return Summoner(id=id, account=account, name=name, region=region)


def get_champion(key: Union[str, int], region: Union[Region, str] = None) -> Champion:
    return get_champions(region=region)[key]


def get_champions(region: Union[Region, str] = None) -> List[Champion]:
    return Champions(region=region)


def get_masteries(region: Union[Region, str] = None) -> List[Mastery]:
    return Masteries(region=region)


def get_runes(region: Union[Region, str] = None) -> List[Rune]:
    return Runes(region=region)


def get_summoner_spells(region: Union[Region, str] = None) -> List[SummonerSpell]:
    return SummonerSpells(region=region)


def get_items(region: Union[Region, str] = None) -> List[Item]:
    return Items(region=region)


def get_mastery_pages(summoner: Summoner, region: Union[Region, str] = None) -> List[MasteryPage]:
    return MasteryPages(summoner=summoner, region=region)


def get_rune_pages(summoner: Summoner, region: Union[Region, str] = None) -> List[RunePage]:
    return RunePages(summoner=summoner, region=region)


def get_maps(region: Union[Region, str] = None) -> List[Map]:
    return Maps(region=region)


def get_profile_icons(region: Union[Region, str] = None) -> List[ProfileIcon]:
    return ProfileIcons(region=region)


def get_realms(region: Union[Region, str] = None) -> Realms:
    return Realms(region=region)


def get_status(region: Union[Region, str] = None) -> ShardStatus:
    return ShardStatus(region=region)


def get_language_strings(region: Union[Region, str] = None) -> LanguageStrings:
    return LanguageStrings(region=region)


def get_locales(region: Union[Region, str] = None) -> List[str]:
    return Locales(region=region)


def get_versions(region: Union[Region, str] = None) -> List[str]:
    return Versions(region=region)


def get_version(date: datetime.date = None, region: Union[Region, str] = None) -> Union[None, str]:
    versions = get_versions(region)
    if date is None:
        return versions[0]
    else:
        patch = get_patch(date)
        for version in versions:
            if patch.majorminor in version:
                return version
    return None


def get_patch(date: datetime.date) -> Patch:
    for _date in PATCHES:
        if _date >= date:
            return PATCHES[_date]
    return PATCHES[next(reversed(PATCHES))]
